def test_cross_repo_search_migration(migrate):
    apps = migrate([("ansible", "0050_crossrepositorycollectionversionindex")])
    AnsibleRepository = apps.get_model("ansible", "AnsibleRepository")
    AnsibleDistribution = apps.get_model("ansible", "AnsibleDistribution")

    # Make a Repository
    repository = AnsibleRepository.objects.create(pulp_type="ansible", name="foobar")
    repository.versions.create(number=0)

    # Make a Distribution
    AnsibleDistribution.objects.create(
        pulp_type="ansible", name="foobar", base_path="foobar", repository=repository
    )

    # Migrate the data
    migrate([("ansible", "0051_cvindex_build")])



"""Tests that CRUD remotes."""

import pytest
import uuid

from pulpcore.client.pulp_ansible import ApiException


@pytest.mark.parallel
def test_crud_role_remote(ansible_bindings, ansible_role_remote_factory, monitor_task):
    # Create a remote.
    remote = ansible_role_remote_factory(
        password=str(uuid.uuid4()),
        username=str(uuid.uuid4()),
        policy="immediate",
    )
    # Try to create a second remote with an identical name.
    with pytest.raises(ApiException):
        ansible_role_remote_factory(name=remote.name)
    # Update a remote using HTTP PATCH.
    new_values = {
        "url": "http://" + str(uuid.uuid4()),
        "username": str(uuid.uuid4()),
        "password": str(uuid.uuid4()),
        "policy": "immediate",
    }
    monitor_task(ansible_bindings.RemotesRoleApi.partial_update(remote.pulp_href, new_values).task)
    # Test read remote
    remote_read = ansible_bindings.RemotesRoleApi.read(remote.pulp_href)
    assert remote_read.name == remote.name
    assert remote_read.url == new_values["url"]
    assert remote_read.policy == new_values["policy"]
    # Update a remote using HTTP PUT.
    new_values = {
        "name": str(uuid.uuid4()),
        "url": "http://" + str(uuid.uuid4()),
        "username": str(uuid.uuid4()),
        "password": str(uuid.uuid4()),
        "policy": "immediate",
    }
    monitor_task(ansible_bindings.RemotesRoleApi.update(remote.pulp_href, new_values).task)
    # Test read remotes
    remotes = ansible_bindings.RemotesRoleApi.list(name=new_values["name"]).results
    assert remotes[0].name == new_values["name"]
    assert remotes[0].url == new_values["url"]
    assert remotes[0].policy == new_values["policy"]
    # Test delete remote
    monitor_task(ansible_bindings.RemotesRoleApi.delete(remote.pulp_href).task)



# Generated by Django 3.2.14 on 2022-07-21 23:05

from django.db import migrations, models
from pulpcore.plugin.migrations import RequireVersion


class Migration(migrations.Migration):

    dependencies = [
        ("ansible", "0058_fix_0056_regression"),
        ("core", "0110_apiappstatus"),
    ]

    operations = [
        RequireVersion("ansible", "0.23.0"),
        migrations.AlterField(
            model_name="collectionversion",
            name="sha256",
            field=models.CharField(db_index=True, max_length=64, null=False),
        ),
        # ---
        # The previous fix in migration 0058 did not do the job.
        # But we can just drop the contstraint if present.
        # And this is only necessary if this migration failed to apply before.
        migrations.RunSQL(
            sql="ALTER TABLE ansible_collectionversion DROP CONSTRAINT IF EXISTS "
            "ansible_collectionversion_sha256_a4d120ab_uniq",
            reverse_sql="",
            elidable=True,
        ),
        # ---
        migrations.AlterUniqueTogether(
            name="collectionversion",
            unique_together={("sha256",)},
        ),
        migrations.RemoveConstraint(
            model_name="collectionversion",
            name="unique_sha256",
        ),
    ]



# Generated by Django 3.2.11 on 2022-02-25 20:39

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('ansible', '0039_collectionremote_signed_only'),
    ]

    operations = [
        migrations.AddField(
            model_name='ansiblerepository',
            name='keyring',
            field=models.FilePathField(blank=True, path='/etc/pulp/certs/', recursive=True),
        ),
    ]



# Generated by Django 4.2.16 on 2024-12-05 12:33

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('ansible', '0035_deprecation_content'),
    ]

    operations = [
        migrations.AlterUniqueTogether(
            name='ansiblecollectiondeprecated',
            unique_together={('namespace', 'name')},
        ),
        migrations.RemoveField(
            model_name='ansiblecollectiondeprecated',
            name='collection_id',
        ),
        migrations.RemoveField(
            model_name='ansiblecollectiondeprecated',
            name='repository_id',
        ),
        migrations.RemoveField(
            model_name='ansiblecollectiondeprecated',
            name='version_added_id',
        ),
    ]



"""Utilities for tests for the ansible plugin."""

import random
import string
from urllib.parse import urlparse, parse_qs


def randstr():
    return "".join(random.choices(string.ascii_lowercase, k=8))


def content_counts(repository_version, summary_type="present"):
    content_summary = getattr(repository_version.content_summary, summary_type)
    return {key: value["count"] for key, value in content_summary.items()}


def iterate_all(list_func, **kwargs):
    """
    Iterate through all of the items on every page in a paginated list view.
    """
    kwargs
    while kwargs is not None:
        response = list_func(**kwargs)

        for x in response.results:
            yield x

        if response.next:
            qs = parse_qs(urlparse(response.next).query)
            for param in ("offset", "limit"):
                if param in qs:
                    kwargs[param] = qs[param][0]
        else:
            kwargs = None



"""Tests that Collections hosted by Pulp can be installed by ansible-galaxy."""

import subprocess
import pytest
from pulpcore.client.pulp_ansible import AnsibleRepositorySyncURL

from pulp_ansible.tests.functional.constants import (
    ANSIBLE_DEMO_COLLECTION,
    ANSIBLE_DEMO_COLLECTION_VERSION,
    GALAXY_ANSIBLE_BASE_URL,
)


@pytest.fixture
def install_scenario_distribution(
    delete_orphans_pre,
    ansible_bindings,
    ansible_repo_factory,
    ansible_collection_remote_factory,
    ansible_distribution_factory,
    monitor_task,
):
    """Prepare a distribution to install from."""
    collection_requirements = f"collections:\n  - {ANSIBLE_DEMO_COLLECTION}"
    remote = ansible_collection_remote_factory(
        url=GALAXY_ANSIBLE_BASE_URL, requirements_file=collection_requirements
    )
    repo = ansible_repo_factory(remote=remote.pulp_href)

    # Sync
    repository_sync_data = AnsibleRepositorySyncURL()
    sync_response = ansible_bindings.RepositoriesAnsibleApi.sync(
        repo.pulp_href, repository_sync_data
    )
    monitor_task(sync_response.task)

    # Distribute
    return ansible_distribution_factory(repo)


def get_current_download_count(api_client, path, namespace, name):
    return int(
        api_client.read(
            path=path,
            namespace=namespace,
            name=name,
        ).download_count
    )


def test_collection_download_count(
    ansible_bindings,
    ansible_dir_factory,
    install_scenario_distribution,
    pulp_settings,
):
    """Test that the collection download counting functions."""
    if not pulp_settings.ANSIBLE_COLLECT_DOWNLOAD_COUNT:
        pytest.skip("ANSIBLE_COLLECT_DOWNLOAD_COUNT not enabled")

    collection_namespace = ANSIBLE_DEMO_COLLECTION.split(".")[0]
    collection_name = ANSIBLE_DEMO_COLLECTION.split(".")[1]
    download_count = get_current_download_count(
        api_client=ansible_bindings.PulpAnsibleApiV3CollectionsApi,
        path=install_scenario_distribution.base_path,
        namespace=collection_namespace,
        name=collection_name,
    )

    temp_dir = ansible_dir_factory(install_scenario_distribution.client_url)

    cmd = [
        "ansible-galaxy",
        "collection",
        "install",
        f"{ANSIBLE_DEMO_COLLECTION}:{ANSIBLE_DEMO_COLLECTION_VERSION}",
        "-c",
        "-p",
        temp_dir,
    ]

    directory = temp_dir / "ansible_collections" / ANSIBLE_DEMO_COLLECTION.replace(".", "/")

    assert not directory.exists(), "Directory {} already exists".format(directory)
    subprocess.run(cmd, cwd=temp_dir)
    assert directory.exists(), "Could not find directory {}".format(directory)

    assert (download_count + 1) == get_current_download_count(
        api_client=ansible_bindings.PulpAnsibleApiV3CollectionsApi,
        path=install_scenario_distribution.base_path,
        namespace=collection_namespace,
        name=collection_name,
    )

    # TODO update to use a second collection version when available
    cmd = [
        "ansible-galaxy",
        "collection",
        "install",
        f"{ANSIBLE_DEMO_COLLECTION}:{ANSIBLE_DEMO_COLLECTION_VERSION}",
        "-c",
        "-f",
        "-p",
        temp_dir,
    ]

    subprocess.run(cmd, cwd=temp_dir)

    assert (download_count + 2) == get_current_download_count(
        api_client=ansible_bindings.PulpAnsibleApiV3CollectionsApi,
        path=install_scenario_distribution.base_path,
        namespace=collection_namespace,
        name=collection_name,
    )



# Generated by Django 3.2.18 on 2023-03-20 13:08

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ('core', '0102_add_domain_relations'),
        ('ansible', '0049_rbac_permissions'),
    ]

    operations = [
        migrations.CreateModel(
            name='CrossRepositoryCollectionVersionIndex',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('is_deprecated', models.BooleanField()),
                ('is_signed', models.BooleanField()),
                ('is_highest', models.BooleanField()),
                ('collection_version', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='ansible_crossrepositorycollectionversionindex', to='ansible.collectionversion')),
                ('namespace_metadata', models.ForeignKey(null=True, on_delete=django.db.models.deletion.SET_NULL, related_name='ansible_crossrepositorycollectionversionindex', to='ansible.ansiblenamespacemetadata')),
                ('repository', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='ansible_crossrepositorycollectionversionindex', to='ansible.ansiblerepository')),
                ('repository_version', models.ForeignKey(null=True, on_delete=django.db.models.deletion.CASCADE, related_name='ansible_crossrepositorycollectionversionindex', to='core.repositoryversion')),
            ],
            options={
                'default_related_name': '%(app_label)s_%(model_name)s',
                'unique_together': {('repository', 'repository_version', 'collection_version')},
            },
        ),
    ]



# Generated by Django 3.2.11 on 2022-01-10 21:26

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('ansible', '0038_collectionversionsignature'),
    ]

    operations = [
        migrations.AddField(
            model_name='collectionremote',
            name='signed_only',
            field=models.BooleanField(default=False),
        ),
    ]



from django.conf import settings
from rest_framework import serializers

from pulp_ansible.app.models import Role


class GalaxyRoleSerializer(serializers.ModelSerializer):
    """
    A serializer for Galaxy's representation of Roles.
    """

    id = serializers.SerializerMethodField(read_only=True)
    name = serializers.CharField()
    namespace = serializers.CharField()

    def get_id(self, obj) -> str:
        """
        Get id.
        """
        return "{}.{}".format(obj.namespace, obj.name)

    class Meta:
        fields = ("id", "name", "namespace")
        model = Role


class GalaxyRoleVersionSerializer(serializers.Serializer):
    """
    A serializer for Galaxy's representation of Role versions.
    """

    name = serializers.CharField(source="version")

    source = serializers.SerializerMethodField(read_only=True)

    def get_source(self, obj) -> str:
        """
        Get source.
        """
        distro_base = self.context["path"]
        distro_path = "".join([settings.CONTENT_ORIGIN, settings.CONTENT_PATH_PREFIX, distro_base])

        return "".join([distro_path, "/", obj.relative_path])

    class Meta:
        fields = ("name", "source")
        model = Role



# Generated by Django 4.2.16 on 2024-12-05 13:18

import django.contrib.postgres.fields
import django.contrib.postgres.search
from django.db import migrations, models
import django.db.models.deletion
import uuid


class Migration(migrations.Migration):

    replaces = [
        ("ansible", "0001_initial"),
        ("ansible", "0002_advanced_collections"),
        ("ansible", "0003_add_tags_and_collectionversion_fields"),
        ("ansible", "0004_add_fulltext_search_indexes"),
        ("ansible", "0005_collectionversion_is_highest"),
        ("ansible", "0006_remove_whitelist_and_alter_collection_version_name"),
        ("ansible", "0007_collectionversion_is_certified"),
        ("ansible", "0008_collectionremote_requirements_file"),
        ("ansible", "0009_collectionimport"),
        ("ansible", "0010_ansible_related_names"),
        ("ansible", "0011_collectionimport"),
        ("ansible", "0012_auto_20190906_2253"),
        ("ansible", "0013_pulp_fields"),
        ("ansible", "0014_certification_enum"),
        ("ansible", "0015_ansiblerepository"),
        ("ansible", "0016_add_extension"),
        ("ansible", "0017_increase_length_collectionversion_fields"),
        ("ansible", "0018_fix_collection_relative_path"),
        ("ansible", "0019_collection_token"),
        ("ansible", "0020_auto_20200810_1926"),
        ("ansible", "0021_rename_role_remote"),
        ("ansible", "0022_URLField_to_CharField"),
        ("ansible", "0023_alter_requirements_file_field"),
        ("ansible", "0024_remove_collectionversion_certification"),
        ("ansible", "0025_increase_collection_version_version_size"),
        ("ansible", "0026_deprecation_per_repository"),
        ("ansible", "0027_tag_length"),
        ("ansible", "0028_collectionversion_namespace_length"),
        ("ansible", "0029_manifest_and_files_json_fields"),
        ("ansible", "0030_collectionversion_requires_ansible"),
        ("ansible", "0031_ansiblerepository_last_synced_metadata_time"),
        ("ansible", "0032_collectionremote_sync_dependencies"),
        ("ansible", "0033_swap_distribution_model"),
        ("ansible", "0034_handle_jsonfield_warnings"),
        ("ansible", "0035_deprecation_content"),
        ("ansible", "0036_update_repository_content"),
        ("ansible", "0037_gitremote"),
        ("ansible", "0038_collectionversionsignature"),
        ("ansible", "0039_collectionremote_signed_only"),
        ("ansible", "0040_ansiblerepository_keyring"),
    ]

    dependencies = [
        ("core", "0091_systemid"),
    ]

    operations = [
        migrations.CreateModel(
            name="Collection",
            fields=[
                (
                    "pulp_id",
                    models.UUIDField(
                        default=uuid.uuid4, editable=False, primary_key=True, serialize=False
                    ),
                ),
                ("pulp_created", models.DateTimeField(auto_now_add=True)),
                ("pulp_last_updated", models.DateTimeField(auto_now=True, null=True)),
                ("namespace", models.CharField(editable=False, max_length=64)),
                ("name", models.CharField(editable=False, max_length=64)),
            ],
            options={
                "unique_together": {("namespace", "name")},
            },
        ),
        migrations.CreateModel(
            name="Tag",
            fields=[
                (
                    "pulp_id",
                    models.UUIDField(
                        default=uuid.uuid4, editable=False, primary_key=True, serialize=False
                    ),
                ),
                ("pulp_created", models.DateTimeField(auto_now_add=True)),
                ("pulp_last_updated", models.DateTimeField(auto_now=True, null=True)),
                ("name", models.CharField(editable=False, max_length=64, unique=True)),
            ],
            options={
                "abstract": False,
            },
        ),
        migrations.CreateModel(
            name="Role",
            fields=[
                (
                    "content_ptr",
                    models.OneToOneField(
                        auto_created=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        parent_link=True,
                        primary_key=True,
                        related_name="ansible_role",
                        serialize=False,
                        to="core.content",
                    ),
                ),
                ("namespace", models.CharField(max_length=64)),
                ("name", models.CharField(max_length=64)),
                ("version", models.CharField(max_length=128)),
            ],
            options={
                "unique_together": {("version", "name", "namespace")},
                "default_related_name": "%(app_label)s_%(model_name)s",
            },
            bases=("core.content",),
        ),
        migrations.CreateModel(
            name="CollectionImport",
            fields=[
                ("messages", models.JSONField(default=list, editable=False)),
                (
                    "task",
                    models.OneToOneField(
                        editable=False,
                        on_delete=django.db.models.deletion.CASCADE,
                        primary_key=True,
                        related_name="+",
                        serialize=False,
                        to="core.task",
                    ),
                ),
            ],
            options={
                "abstract": False,
                "ordering": ["task__pulp_created"],
            },
        ),
        migrations.CreateModel(
            name="RoleRemote",
            fields=[
                (
                    "remote_ptr",
                    models.OneToOneField(
                        auto_created=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        parent_link=True,
                        primary_key=True,
                        related_name="ansible_roleremote",
                        serialize=False,
                        to="core.remote",
                    ),
                ),
            ],
            options={
                "abstract": False,
                "default_related_name": "%(app_label)s_%(model_name)s",
            },
            bases=("core.remote",),
        ),
        migrations.CreateModel(
            name="AnsibleRepository",
            fields=[
                (
                    "repository_ptr",
                    models.OneToOneField(
                        auto_created=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        parent_link=True,
                        primary_key=True,
                        related_name="ansible_ansiblerepository",
                        serialize=False,
                        to="core.repository",
                    ),
                ),
                ("last_synced_metadata_time", models.DateTimeField(null=True)),
                (
                    "keyring",
                    models.FilePathField(blank=True, path="/etc/pulp/certs/", recursive=True),
                ),
            ],
            options={
                "default_related_name": "%(app_label)s_%(model_name)s",
                "permissions": (
                    ("modify_ansible_repo_content", "Can modify ansible repository content"),
                ),
            },
            bases=("core.repository",),
        ),
        migrations.CreateModel(
            name="collectionversion",
            fields=[
                (
                    "content_ptr",
                    models.OneToOneField(
                        auto_created=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        parent_link=True,
                        primary_key=True,
                        related_name="ansible_collectionversion",
                        serialize=False,
                        to="core.content",
                    ),
                ),
                ("version", models.CharField(editable=False, max_length=128)),
                ("contents", models.JSONField(default=list, editable=False)),
                (
                    "collection",
                    models.ForeignKey(
                        editable=False,
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="versions",
                        to="ansible.collection",
                        to_field="pulp_id",
                    ),
                ),
                (
                    "authors",
                    django.contrib.postgres.fields.ArrayField(
                        base_field=models.CharField(max_length=64),
                        default=list,
                        editable=False,
                        size=None,
                    ),
                ),
                ("dependencies", models.JSONField(default=dict, editable=False)),
                ("description", models.TextField(blank=True, default="", editable=False)),
                ("docs_blob", models.JSONField(default=dict, editable=False)),
                (
                    "documentation",
                    models.CharField(blank=True, default="", editable=False, max_length=2000),
                ),
                (
                    "homepage",
                    models.CharField(blank=True, default="", editable=False, max_length=2000),
                ),
                (
                    "issues",
                    models.CharField(blank=True, default="", editable=False, max_length=2000),
                ),
                (
                    "license",
                    django.contrib.postgres.fields.ArrayField(
                        base_field=models.CharField(max_length=32),
                        default=list,
                        editable=False,
                        size=None,
                    ),
                ),
                ("name", models.CharField(editable=False, max_length=64)),
                ("namespace", models.CharField(editable=False, max_length=64)),
                (
                    "repository",
                    models.CharField(blank=True, default="", editable=False, max_length=2000),
                ),
                (
                    "tags",
                    models.ManyToManyField(
                        editable=False, related_name="ansible_collectionversion", to="ansible.tag"
                    ),
                ),
                ("search_vector", django.contrib.postgres.search.SearchVectorField(default="")),
                ("is_highest", models.BooleanField(default=False, editable=False)),
                ("files", models.JSONField(default=dict, editable=False)),
                ("manifest", models.JSONField(default=dict, editable=False)),
                ("requires_ansible", models.CharField(max_length=255, null=True)),
            ],
            options={
                "default_related_name": "%(app_label)s_%(model_name)s",
                "unique_together": {("namespace", "name", "version")},
            },
            bases=("core.content",),
        ),
        migrations.CreateModel(
            name="AnsibleDistribution",
            fields=[
                (
                    "distribution_ptr",
                    models.OneToOneField(
                        auto_created=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        parent_link=True,
                        primary_key=True,
                        related_name="ansible_ansibledistribution",
                        serialize=False,
                        to="core.distribution",
                    ),
                ),
            ],
            options={
                "default_related_name": "%(app_label)s_%(model_name)s",
            },
            bases=("core.distribution",),
        ),
        migrations.CreateModel(
            name="AnsibleCollectionDeprecated",
            fields=[
                (
                    "content_ptr",
                    models.OneToOneField(
                        auto_created=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        parent_link=True,
                        primary_key=True,
                        related_name="ansible_ansiblecollectiondeprecated",
                        serialize=False,
                        to="core.content",
                    ),
                ),
                ("namespace", models.CharField(editable=False, max_length=64)),
                ("name", models.CharField(editable=False, max_length=64)),
            ],
            options={
                "default_related_name": "%(app_label)s_%(model_name)s",
                "unique_together": {("namespace", "name")},
            },
            bases=("core.content",),
        ),
        migrations.CreateModel(
            name="GitRemote",
            fields=[
                (
                    "remote_ptr",
                    models.OneToOneField(
                        auto_created=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        parent_link=True,
                        primary_key=True,
                        related_name="ansible_gitremote",
                        serialize=False,
                        to="core.remote",
                    ),
                ),
                ("metadata_only", models.BooleanField(default=False)),
                ("git_ref", models.TextField()),
            ],
            options={
                "default_related_name": "%(app_label)s_%(model_name)s",
            },
            bases=("core.remote",),
        ),
        migrations.CreateModel(
            name="CollectionRemote",
            fields=[
                (
                    "remote_ptr",
                    models.OneToOneField(
                        auto_created=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        parent_link=True,
                        primary_key=True,
                        related_name="ansible_collectionremote",
                        serialize=False,
                        to="core.remote",
                    ),
                ),
                ("requirements_file", models.TextField(null=True)),
                ("auth_url", models.CharField(max_length=255, null=True)),
                ("token", models.TextField(max_length=2000, null=True)),
                ("sync_dependencies", models.BooleanField(default=True)),
                ("signed_only", models.BooleanField(default=False)),
            ],
            options={
                "abstract": False,
                "default_related_name": "%(app_label)s_%(model_name)s",
            },
            bases=("core.remote",),
        ),
        migrations.CreateModel(
            name="CollectionVersionSignature",
            fields=[
                (
                    "content_ptr",
                    models.OneToOneField(
                        auto_created=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        parent_link=True,
                        primary_key=True,
                        related_name="ansible_collectionversionsignature",
                        serialize=False,
                        to="core.content",
                    ),
                ),
                ("data", models.BinaryField()),
                ("digest", models.CharField(max_length=64)),
                ("pubkey_fingerprint", models.CharField(max_length=64)),
                (
                    "signed_collection",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="signatures",
                        to="ansible.collectionversion",
                    ),
                ),
                (
                    "signing_service",
                    models.ForeignKey(
                        null=True,
                        on_delete=django.db.models.deletion.SET_NULL,
                        related_name="signatures",
                        to="core.signingservice",
                    ),
                ),
            ],
            options={
                "default_related_name": "%(app_label)s_%(model_name)s",
                "unique_together": {("pubkey_fingerprint", "signed_collection")},
            },
            bases=("core.content",),
        ),
        migrations.AddConstraint(
            model_name="collectionversion",
            constraint=models.UniqueConstraint(
                condition=models.Q(("is_highest", True)),
                fields=("collection", "is_highest"),
                name="unique_is_highest",
            ),
        ),
        migrations.RunSQL(
            sql="\nUPDATE ansible_collectionversion AS c\nSET search_vector = (\n    SELECT \n    setweight(to_tsvector(coalesce(namespace,'')), 'A')\n    || setweight(to_tsvector(coalesce(name, '')), 'A')\n    || (\n      SELECT\n        setweight(to_tsvector(\n          coalesce(string_agg(\"ansible_tag\".\"name\", ' '), '')\n        ), 'B')\n      FROM\n        \"ansible_tag\" INNER JOIN \"ansible_collectionversion_tags\" ON (\"ansible_tag\".\"pulp_id\" = \"ansible_collectionversion_tags\".\"tag_id\")\n    )\n    || (\n      SELECT\n        setweight(to_tsvector(\n          coalesce(string_agg(cvc ->> 'name', ' '), '')\n        ), 'C')\n      FROM jsonb_array_elements(cv.contents) AS cvc\n    )\n    || setweight(to_tsvector(coalesce(description, '')), 'D')\n\n    FROM ansible_collectionversion cv\n    WHERE c.content_ptr_id = cv.content_ptr_id\n)\n",
            reverse_sql="",
        ),
        migrations.RunSQL(
            sql="\nCREATE OR REPLACE FUNCTION update_collection_ts_vector()\n    RETURNS TRIGGER AS\n$$\nBEGIN\n    NEW.search_vector := (\n        SELECT \n    setweight(to_tsvector(coalesce(namespace,'')), 'A')\n    || setweight(to_tsvector(coalesce(name, '')), 'A')\n    || (\n      SELECT\n        setweight(to_tsvector(\n          coalesce(string_agg(\"ansible_tag\".\"name\", ' '), '')\n        ), 'B')\n      FROM\n        \"ansible_tag\" INNER JOIN \"ansible_collectionversion_tags\" ON (\"ansible_tag\".\"pulp_id\" = \"ansible_collectionversion_tags\".\"tag_id\")\n    )\n    || (\n      SELECT\n        setweight(to_tsvector(\n          coalesce(string_agg(cvc ->> 'name', ' '), '')\n        ), 'C')\n      FROM jsonb_array_elements(cv.contents) AS cvc\n    )\n    || setweight(to_tsvector(coalesce(description, '')), 'D')\n\n        FROM ansible_collectionversion cv\n        WHERE cv.content_ptr_id = NEW.content_ptr_id\n    );\n    RETURN NEW;\nEND;\n$$ LANGUAGE plpgsql;\nCREATE TRIGGER update_ts_vector\n    BEFORE UPDATE\n    ON ansible_collectionversion\n    FOR EACH ROW\nEXECUTE PROCEDURE update_collection_ts_vector();\n",
            reverse_sql="\nDROP TRIGGER IF EXISTS update_ts_vector ON ansible_collectionversion;\nDROP FUNCTION IF EXISTS update_collection_ts_vector();\n",
        ),
    ]



# Generated by Django 3.2.18 on 2023-03-20 13:08

from django.db import migrations
from django.db.models import Q

from semantic_version import Version


#####################################################
# COPIED FROM tasks/collectionversion_index.py
#####################################################

def has_distribution(repository, repository_version, apps=None):
    """Is -any- distribution attached to this repo or repo version?"""
    AnsibleDistribution = apps.get_model('ansible', 'AnsibleDistribution')
    return AnsibleDistribution.objects.filter(
        Q(repository=repository) | Q(repository_version=repository_version)
    ).exists()


def get_highest_version_string_from_cv_objects(cv_objects):
    """Return the highest version as a string (preferring stable releases)."""
    versions_strings = [x.version for x in cv_objects]
    versions = [Version(x) for x in versions_strings]
    versions = sorted(versions, reverse=True)
    stable_versions = [x for x in versions if not x.prerelease]
    if stable_versions:
        return str(stable_versions[0])
    return str(versions[0])


def get_highest_repository_version(apps, repository):
    RepositoryVersion = apps.get_model('core', 'RepositoryVersion')
    return RepositoryVersion.objects.filter(repository=repository).order_by('number').last()


def get_repository_version_content(apps, repository_version):
    Content = apps.get_model('core', 'Content')
    RepositoryContent = apps.get_model('core', 'RepositoryContent')

    # https://github.com/pulp/pulpcore/blob/main/pulpcore/app/models/repository.py#L660-L662
    content_relationships = RepositoryContent.objects.filter(
        repository=repository_version.repository, version_added__number__lte=repository_version.number
    ).exclude(version_removed__number__lte=repository_version.number)

    # https://github.com/pulp/pulpcore/blob/main/pulpcore/app/models/repository.py#L687
    return Content.objects.filter(version_memberships__in=content_relationships)


def update_index(apps=None, distribution=None, repository=None, repository_version=None, is_latest=False):
    """Rebuild index by distribtion|repository|repositoryversion."""

    # deferred model import ...
    AnsibleCollectionDeprecated = apps.get_model('ansible', 'AnsibleCollectionDeprecated')
    AnsibleNamespace = apps.get_model('ansible', 'AnsibleNamespace')
    AnsibleRepository = apps.get_model('ansible', 'AnsibleRepository')
    CollectionVersion = apps.get_model('ansible', 'CollectionVersion')
    CollectionVersionSignature = apps.get_model('ansible', 'CollectionVersionSignature')
    CVIndex = apps.get_model('ansible', 'CrossRepositoryCollectionVersionIndex')

    # if the distro points at a specific repo version, we should use that in the index
    # otherwise the index value for repository version should be null
    # use_repository_version = False
    use_repository_version = not is_latest

    # a repov was passed in so we should use that
    # if repository_version:
    #    use_repository_version = True

    # make sure the distro points at a repo[version]
    if distribution and not repository and not repository_version:
        # sometimes distros point at a version
        if distribution.repository_version:
            repository = distribution.repository_version.repository
            repository_version = distribution.repository_version
            use_repository_version = True

        # sometimes distros point at a repository
        elif distribution.repository is not None:
            repository = distribution.repository
            if hasattr(distribution.repository, 'latest_version'):
                repository_version = distribution.repository.latest_version()
            else:
                repository_version = get_highest_repository_version(apps, distribution.repository)
            use_repository_version = False

        # sometimes distros point at nothing?
        else:
            return

    # extract repository from repository version if needed
    if repository is None:
        repository = repository_version.repository

    # optimization: -must- have an AnsibleRepository for the index model
    if not isinstance(repository, AnsibleRepository):
        repository = AnsibleRepository.objects.filter(pk=repository.pk).first()
        if repository is None:
            return

    # optimization: we only want to index "distributed" CVs
    if distribution is None and not has_distribution(repository, repository_version, apps=apps):
        return

    # This block handles a case where the distribution that points at a repository
    # has been deleted. If no other distribution points at the repository, all related
    # indexes need to be removed and to exit early.
    if not use_repository_version:
        if not has_distribution(repository, repository_version, apps=apps):
            CVIndex.objects.filter(repository=repository, repository_version=None).delete()
            return

    # optimizaion: exit early if using a repo version and it's alreay indexed
    if use_repository_version:
        if CVIndex.objects.filter(repository_version=repository_version).exists():
            return

    # get all CVs in this repository version
    if hasattr(repository_version, 'content'):
        cvs = repository_version.content.filter(pulp_type="ansible.collection_version").values_list(
            "pk", flat=True
        )
    else:
        repov_content = get_repository_version_content(apps, repository_version)
        cvs = repov_content.filter(pulp_type="ansible.collection_version").values_list(
            "pk", flat=True
        )

    cvs = CollectionVersion.objects.filter(pk__in=cvs)

    # clean out cvs no longer in the repo when a distro w/ a repo
    if not use_repository_version:
        CVIndex.objects.filter(repository=repository, repository_version=None).exclude(
            collection_version__pk__in=cvs
        ).delete()

    # get the set of signatures in this repo version
    if hasattr(repository_version, 'content'):
        repo_signatures_pks = repository_version.content.filter(
            pulp_type="ansible.collection_signature"
        ).values_list("pk", flat=True)
    else:
        repov_content = get_repository_version_content(apps, repository_version)
        repo_signatures_pks = repov_content.filter(
            pulp_type="ansible.collection_signature"
        ).values_list("pk", flat=True)
    repo_signatures = CollectionVersionSignature.objects.filter(pk__in=repo_signatures_pks)

    # get the set of deprecations in this repo version
    if hasattr(repository_version, 'content'):
        deprecations = repository_version.content.filter(
            pulp_type="ansible.collection_deprecation"
        ).values_list("pk", flat=True)
    else:
        repov_content = get_repository_version_content(apps, repository_version)
        deprecations = repov_content.filter(
            pulp_type="ansible.collection_deprecation"
        ).values_list("pk", flat=True)
    deprecations = AnsibleCollectionDeprecated.objects.filter(pk__in=deprecations)
    deprecations_set = {(x.namespace, x.name) for x in deprecations}

    # find all namespaces in the repo version
    if hasattr(repository_version, 'content'):
        namespaces = repository_version.content.filter(pulp_type="ansible.namespace").values_list(
            "pk", flat=True
        )
    else:
        repov_content = get_repository_version_content(apps, repository_version)
        namespaces = repov_content.filter(pulp_type="ansible.namespace").values_list(
            "pk", flat=True
        )
    namespaces = AnsibleNamespace.objects.filter(pk__in=namespaces)
    namespaces = {x.name: x for x in namespaces}

    # map out the namespace(s).name(s) for everything in the repo version
    colset = set(cvs.values_list("namespace", "name").distinct())

    repo_v = None
    if use_repository_version:
        repo_v = repository_version

    # iterate through each collection in the repository
    for colkey in colset:
        namespace, name = colkey

        # get all the versions for this collection
        related_cvs = cvs.filter(namespace=namespace, name=name).only("version")

        # what is the "highest" version in this list?
        highest_version = get_highest_version_string_from_cv_objects(related_cvs)

        # should all of these CVs be deprecated?
        is_deprecated = colkey in deprecations_set

        # process each related CV
        for rcv in related_cvs:
            # get the related signatures for this CV
            rcv_signatures = repo_signatures.filter(signed_collection=rcv).count()

            # create|update the index for this CV
            CVIndex.objects.update_or_create(
                repository=repository,
                repository_version=repo_v,
                collection_version=rcv,
                defaults={
                    "is_highest": rcv.version == highest_version,
                    "is_signed": rcv_signatures > 0,
                    "is_deprecated": is_deprecated,
                    "namespace_metadata": namespaces.get(namespace, None),
                },
            )


def rebuild_index(apps):
    """Rebuild -everything-."""

    indexed_repos = set()

    AnsibleDistribution = apps.get_model('ansible', 'AnsibleDistribution')
    dqs = AnsibleDistribution.objects.select_related(
        "repository", "repository_version", "repository_version__repository"
    ).all()
    for distro in dqs:
        if distro.repository_version:
            rv = distro.repository_version
        else:
            if hasattr(distro.repository, 'latest_version'):
                rv = distro.repository.latest_version()
            else:
                rv = get_highest_repository_version(apps, distro.repository)

        if rv.pk in indexed_repos:
            continue

        update_index(distribution=distro, apps=apps)
        indexed_repos.add(rv.pk)


#####################################################
# MIGRATION CODE
#####################################################

def force_build(apps, *args, **kwargs):
    rebuild_index(apps)


class Migration(migrations.Migration):

    dependencies = [
        ('core', '0102_add_domain_relations'),
        ('ansible', '0050_crossrepositorycollectionversionindex'),
    ]

    operations = [
       migrations.RunPython(
            code=force_build,
            reverse_code=migrations.RunPython.noop,
            elidable=True,
       ),
    ]



from django.conf import settings
from django.utils.module_loading import import_string
from pulpcore.plugin.tasking import dispatch

from pulp_ansible.app.tasks.collections import import_collection


class UploadGalaxyCollectionMixin:
    """
    Provides a method that dispatches a task.
    """

    def _dispatch_import_collection_task(self, temp_file_pk, repository=None, **kwargs):
        """
        Dispatch a Import Collection creation task.
        """
        locks = []
        kwargs["temp_file_pk"] = temp_file_pk
        if repository:
            locks.append(repository)
            kwargs["repository_pk"] = repository.pk

        return dispatch(import_collection, exclusive_resources=locks, kwargs=kwargs)

    def get_deferred_context(self, request):
        context = {}
        if "file" in request.data:
            context["filename"] = request.data["file"].name
        return context


def perform_import(value):
    if isinstance(value, str):
        return import_string(value)
    elif isinstance(value, (tuple, list)):
        return [perform_import(v) for v in value]
    return value


class GalaxyAuthMixin:
    """
    Provides the authentication and permission classes from settings.
    """

    authentication_classes = perform_import(settings.ANSIBLE_AUTHENTICATION_CLASSES)
    permission_classes = perform_import(settings.ANSIBLE_PERMISSION_CLASSES)



# Generated by Django 4.2.19 on 2025-05-21 12:04

import django.contrib.postgres.fields
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("ansible", "0060_remove_collectionversion_unique_is_highest_and_more"),
    ]

    operations = [
        migrations.AddField(
            model_name="collectionversion",
            name="new_tags",
            field=django.contrib.postgres.fields.ArrayField(
                base_field=models.CharField(max_length=64, unique=True),
                db_column="tags",
                null=True,
                size=None,
            ),
        ),
    ]



"""Tests functionality around Collection-Version Signatures."""

import tarfile

import pytest

from pulpcore.tests.functional.utils import PulpTaskError

from pulp_ansible.tests.functional.utils import content_counts
from pulpcore.client.pulp_ansible import AnsibleRepositorySyncURL


def test_upload_then_sign_then_try_to_upload_duplicate_signature(
    ansible_bindings,
    build_and_upload_collection,
    ansible_repo_factory,
    ascii_armored_detached_signing_service,
    tmp_path,
    sign_with_ascii_armored_detached_signing_service,
    monitor_task,
):
    """Test server signing of collections, and that a duplicate signature can't be uploaded."""
    repository = ansible_repo_factory()
    collection, collection_url = build_and_upload_collection()

    # Add it to a repo version
    body = {"add_content_units": [collection_url]}
    monitor_task(ansible_bindings.RepositoriesAnsibleApi.modify(repository.pulp_href, body).task)
    repository = ansible_bindings.RepositoriesAnsibleApi.read(repository.pulp_href)

    # Assert the content was added correctly
    assert repository.latest_version_href.endswith("/1/")

    repository_version = ansible_bindings.RepositoriesAnsibleVersionsApi.read(
        repository.latest_version_href
    )
    assert content_counts(repository_version) == {"ansible.collection_version": 1}

    # Ask Pulp to sign the collection
    body = {
        "content_units": [collection_url],
        "signing_service": ascii_armored_detached_signing_service.pulp_href,
    }
    monitor_task(ansible_bindings.RepositoriesAnsibleApi.sign(repository.pulp_href, body).task)
    repository = ansible_bindings.RepositoriesAnsibleApi.read(repository.pulp_href)

    # Assert a new repo version with signatures was created
    assert repository.latest_version_href.endswith("/2/")
    repository_version = ansible_bindings.RepositoriesAnsibleVersionsApi.read(
        repository.latest_version_href
    )
    assert content_counts(repository_version) == {
        "ansible.collection_signature": 1,
        "ansible.collection_version": 1,
    }

    # Extract MANIFEST.json
    with tarfile.open(collection.filename, mode="r") as tar:
        filename = tmp_path / "MANIFEST.json"
        tar.extract("MANIFEST.json", path=tmp_path)

    # Locally sign the collection
    signing_script_response = sign_with_ascii_armored_detached_signing_service(filename)
    signature = signing_script_response["signature"]

    # Check that locally produced signature can't be uploaded
    with pytest.raises(PulpTaskError):
        task = ansible_bindings.ContentCollectionSignaturesApi.create(
            signed_collection=collection_url, file=signature
        )
        monitor_task(task.task)


def test_sign_locally_then_upload_signature(
    ansible_bindings,
    build_and_upload_collection,
    sign_with_ascii_armored_detached_signing_service,
    tmp_path,
    ansible_repo_factory,
    pulp_trusted_public_key_fingerprint,
    pulp_trusted_public_key,
    monitor_task,
):
    """Test uploading a locally produced Collection Signature."""
    repository = ansible_repo_factory(gpgkey=pulp_trusted_public_key)
    collection, collection_url = build_and_upload_collection()

    # Extract MANIFEST.json
    with tarfile.open(collection.filename, mode="r") as tar:
        filename = tmp_path / "MANIFEST.json"
        tar.extract("MANIFEST.json", path=tmp_path)

    # Locally sign the collection
    signing_script_response = sign_with_ascii_armored_detached_signing_service(filename)
    signature_file = signing_script_response["signature"]

    # Signature upload
    task = ansible_bindings.ContentCollectionSignaturesApi.create(
        signed_collection=collection_url, file=signature_file, repository=repository.pulp_href
    )
    signature_href = next(
        (
            item
            for item in monitor_task(task.task).created_resources
            if "content/ansible/collection_signatures/" in item
        )
    )
    assert signature_href is not None
    signature = ansible_bindings.ContentCollectionSignaturesApi.read(signature_href)
    assert signature.pubkey_fingerprint == pulp_trusted_public_key_fingerprint

    # Upload another collection that won't be signed
    another_collection, another_collection_url = build_and_upload_collection()

    # Check that invalid signatures can't be uploaded
    with pytest.raises(Exception):
        task = ansible_bindings.ContentCollectionSignaturesApi.create(
            signed_collection=another_collection_url,
            file=signature,
            repository=repository.pulp_href,
        )
        monitor_task(task.task)


@pytest.mark.parallel
def test_collection_signatures_immutable(ansible_bindings):
    """Test the API doesn't provide an update or partial_update for a Collection Signature objs."""
    with pytest.raises(AttributeError):
        ansible_bindings.ContentCollectionSignaturesApi.update

    with pytest.raises(AttributeError):
        ansible_bindings.ContentCollectionSignaturesApi.partial_update


@pytest.mark.parallel
def test_collection_signatures_no_deletion(ansible_bindings):
    """Test the API doesn't provide a delete for a Collection Signature obj."""
    with pytest.raises(AttributeError):
        ansible_bindings.ContentCollectionSignaturesApi.delete


# Tests for syncing Collection Signatures


@pytest.fixture
def distro_serving_one_signed_one_unsigned_collection(
    ansible_bindings,
    build_and_upload_collection,
    ascii_armored_detached_signing_service,
    ansible_repo_factory,
    ansible_distribution_factory,
    monitor_task,
):
    """Create a distro serving two collections, one signed, one unsigned."""
    repository = ansible_repo_factory()
    collections = []
    for i in range(2):
        _, collection_url = build_and_upload_collection()
        collections.append(collection_url)

    body = {"add_content_units": collections}
    monitor_task(ansible_bindings.RepositoriesAnsibleApi.modify(repository.pulp_href, body).task)

    body = {
        "content_units": collections[:1],
        "signing_service": ascii_armored_detached_signing_service.pulp_href,
    }
    monitor_task(ansible_bindings.RepositoriesAnsibleApi.sign(repository.pulp_href, body).task)

    return ansible_distribution_factory(repository=repository)


def test_sync_signatures(
    ansible_bindings,
    ansible_repo_factory,
    ansible_collection_remote_factory,
    distro_serving_one_signed_one_unsigned_collection,
    monitor_task,
):
    """Test that signatures are also synced."""
    distro = distro_serving_one_signed_one_unsigned_collection
    repository = ansible_repo_factory()

    # Create Remote
    remote = ansible_collection_remote_factory(url=distro.client_url, include_pulp_auth=True)

    # Sync
    repository_sync_data = AnsibleRepositorySyncURL(remote=remote.pulp_href)
    sync_response = ansible_bindings.RepositoriesAnsibleApi.sync(
        repository.pulp_href, repository_sync_data
    )
    monitor_task(sync_response.task)
    repository = ansible_bindings.RepositoriesAnsibleApi.read(repository.pulp_href)

    repository_version = ansible_bindings.RepositoriesAnsibleVersionsApi.read(
        repository.latest_version_href
    )
    assert content_counts(repository_version) == {
        "ansible.collection_version": 2,
        "ansible.collection_signature": 1,
    }


def test_sync_signatures_only(
    ansible_bindings,
    ansible_repo_factory,
    ansible_collection_remote_factory,
    distro_serving_one_signed_one_unsigned_collection,
    monitor_task,
):
    """Test that only collections with a signatures are synced when specified."""
    distro = distro_serving_one_signed_one_unsigned_collection
    repository = ansible_repo_factory()

    # Create Remote
    remote = ansible_collection_remote_factory(
        url=distro.client_url, signed_only=True, include_pulp_auth=True
    )

    # Sync
    repository_sync_data = AnsibleRepositorySyncURL(remote=remote.pulp_href)
    sync_response = ansible_bindings.RepositoriesAnsibleApi.sync(
        repository.pulp_href, repository_sync_data
    )
    monitor_task(sync_response.task)
    repository = ansible_bindings.RepositoriesAnsibleApi.read(repository.pulp_href)

    repository_version = ansible_bindings.RepositoriesAnsibleVersionsApi.read(
        repository.latest_version_href
    )
    assert content_counts(repository_version) == {
        "ansible.collection_version": 1,
        "ansible.collection_signature": 1,
    }



# Generated by Django 4.2.19 on 2025-04-15 16:27

from django.db import migrations, models
import django.db.models.deletion
import pulpcore.app.util


class Migration(migrations.Migration):

    dependencies = [
        ("core", "0117_task_unblocked_at"),
        ("ansible", "0062_pivot_collectionversion_tags"),
    ]

    operations = [
        migrations.AlterUniqueTogether(
            name="ansiblecollectiondeprecated",
            unique_together=set(),
        ),
        migrations.AlterUniqueTogether(
            name="ansiblenamespacemetadata",
            unique_together=set(),
        ),
        migrations.AlterUniqueTogether(
            name="collectionversion",
            unique_together=set(),
        ),
        migrations.AlterUniqueTogether(
            name="collectionversionmark",
            unique_together=set(),
        ),
        migrations.AlterUniqueTogether(
            name="collectionversionsignature",
            unique_together=set(),
        ),
        migrations.AlterUniqueTogether(
            name="role",
            unique_together=set(),
        ),
        migrations.AddField(
            model_name="ansiblecollectiondeprecated",
            name="_pulp_domain",
            field=models.ForeignKey(
                default=pulpcore.app.util.get_domain_pk,
                on_delete=django.db.models.deletion.PROTECT,
                to="core.domain",
            ),
        ),
        migrations.AddField(
            model_name="ansiblenamespacemetadata",
            name="_pulp_domain",
            field=models.ForeignKey(
                default=pulpcore.app.util.get_domain_pk,
                on_delete=django.db.models.deletion.PROTECT,
                to="core.domain",
            ),
        ),
        migrations.AddField(
            model_name="collectionversion",
            name="_pulp_domain",
            field=models.ForeignKey(
                default=pulpcore.app.util.get_domain_pk,
                on_delete=django.db.models.deletion.PROTECT,
                to="core.domain",
            ),
        ),
        migrations.AddField(
            model_name="collectionversionmark",
            name="_pulp_domain",
            field=models.ForeignKey(
                default=pulpcore.app.util.get_domain_pk,
                on_delete=django.db.models.deletion.PROTECT,
                to="core.domain",
            ),
        ),
        migrations.AddField(
            model_name="collectionversionsignature",
            name="_pulp_domain",
            field=models.ForeignKey(
                default=pulpcore.app.util.get_domain_pk,
                on_delete=django.db.models.deletion.PROTECT,
                to="core.domain",
            ),
        ),
        migrations.AddField(
            model_name="role",
            name="_pulp_domain",
            field=models.ForeignKey(
                default=pulpcore.app.util.get_domain_pk,
                on_delete=django.db.models.deletion.PROTECT,
                to="core.domain",
            ),
        ),
        migrations.AlterUniqueTogether(
            name="ansiblecollectiondeprecated",
            unique_together={("namespace", "name", "_pulp_domain")},
        ),
        migrations.AlterUniqueTogether(
            name="ansiblenamespacemetadata",
            unique_together={("metadata_sha256", "_pulp_domain")},
        ),
        migrations.AlterUniqueTogether(
            name="collectionversion",
            unique_together={("sha256", "_pulp_domain")},
        ),
        migrations.AlterUniqueTogether(
            name="collectionversionmark",
            unique_together={("value", "marked_collection", "_pulp_domain")},
        ),
        migrations.AlterUniqueTogether(
            name="collectionversionsignature",
            unique_together={("pubkey_fingerprint", "signed_collection", "_pulp_domain")},
        ),
        migrations.AlterUniqueTogether(
            name="role",
            unique_together={("version", "name", "namespace", "_pulp_domain")},
        ),
    ]



# Generated by Django 2.2.19 on 2021-03-19 17:42

from django.db import migrations, models, transaction
import django.db.models.deletion


def migrate_data_from_old_model_to_new_model_up(apps, schema_editor):
    """Move objects from AnsibleDistribution to NewAnsibleDistribution."""
    AnsibleDistribution = apps.get_model("ansible", "AnsibleDistribution")
    NewAnsibleDistribution = apps.get_model("ansible", "NewAnsibleDistribution")
    for ansible_distribution in AnsibleDistribution.objects.all():
        with transaction.atomic():
            NewAnsibleDistribution(
                pulp_id=ansible_distribution.pulp_id,
                pulp_created=ansible_distribution.pulp_created,
                pulp_last_updated=ansible_distribution.pulp_last_updated,
                pulp_type=ansible_distribution.pulp_type,
                name=ansible_distribution.name,
                base_path=ansible_distribution.base_path,
                content_guard=ansible_distribution.content_guard,
                remote=ansible_distribution.remote,
                repository_version=ansible_distribution.repository_version,
                repository=ansible_distribution.repository,
            ).save()
            ansible_distribution.delete()


def migrate_data_from_old_model_to_new_model_down(apps, schema_editor):
    """Move objects from NewAnsibleDistribution to AnsibleDistribution."""
    AnsibleDistribution = apps.get_model("ansible", "AnsibleDistribution")
    NewAnsibleDistribution = apps.get_model("ansible", "NewAnsibleDistribution")
    for ansible_distribution in NewAnsibleDistribution.objects.all():
        with transaction.atomic():
            AnsibleDistribution(
                pulp_id=ansible_distribution.pulp_id,
                pulp_created=ansible_distribution.pulp_created,
                pulp_last_updated=ansible_distribution.pulp_last_updated,
                pulp_type=ansible_distribution.pulp_type,
                name=ansible_distribution.name,
                base_path=ansible_distribution.base_path,
                content_guard=ansible_distribution.content_guard,
                remote=ansible_distribution.remote,
                repository_version=ansible_distribution.repository_version,
                repository=ansible_distribution.repository,
            ).save()
            ansible_distribution.delete()


class Migration(migrations.Migration):
    atomic = False

    dependencies = [
        ("ansible", "0032_collectionremote_sync_dependencies"),
    ]

    operations = [
        migrations.CreateModel(
            name="NewAnsibleDistribution",
            fields=[
                (
                    "distribution_ptr",
                    models.OneToOneField(
                        auto_created=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        parent_link=True,
                        primary_key=True,
                        related_name="ansible_ansibledistribution",
                        serialize=False,
                        to="core.Distribution",
                    ),
                ),
            ],
            options={
                "default_related_name": "%(app_label)s_%(model_name)s",
            },
            bases=("core.distribution",),
        ),
        migrations.RunPython(
            code=migrate_data_from_old_model_to_new_model_up,
            reverse_code=migrate_data_from_old_model_to_new_model_down,
            elidable=True,
        ),
        migrations.DeleteModel(
            name="AnsibleDistribution",
        ),
        migrations.RenameModel(
            old_name="NewAnsibleDistribution",
            new_name="AnsibleDistribution",
        ),
    ]



from drf_spectacular.utils import extend_schema_field
from drf_spectacular.types import OpenApiTypes
from rest_framework import serializers


@extend_schema_field(OpenApiTypes.OBJECT)
class JSONDictField(serializers.JSONField):
    """A drf JSONField override to force openapi schema to use 'object' type.

    Not strictly correct, but we relied on that for a long time.
    See: https://github.com/tfranzel/drf-spectacular/issues/1095
    """


@extend_schema_field({"type": "array", "items": {}})
class JSONListField(serializers.JSONField):
    """A drf JSONField override to force openapi schema to use 'array' type."""



"""
This module includes tasks related to deleting Content.

You can remove Content from a Repository by making a new RepositoryVersion
without the Content. If an API endpoint uses a Distribution which points to
the latest_version of the Repository then the Content is unavailable,
however it is not deleted.

Content can only be deleted if it exists in no RepositoryVersion.

Content cannot be removed from a RepositoryVersion since it is immutable.

Pulp's orphan_cleanup task deletes any Content not part of a RepositoryVersion.
"""

import logging
from collections import defaultdict

from pulp_ansible.app.models import Collection, CollectionVersion
from pulpcore.plugin.tasking import add_and_remove, orphan_cleanup

log = logging.getLogger(__name__)


def _cleanup_old_versions(repo):
    """Delete all the old versions of the given repository."""
    for version in repo.versions.complete().order_by("-number")[1:]:
        version.delete()


def _remove_collection_version_from_repos(collection_versions):
    """Remove CollectionVersions from latest RepositoryVersion of each repo."""
    repos_with_collections_to_delete = defaultdict(list)
    for collection_version in collection_versions:
        for repo in collection_version.repositories.all():
            repos_with_collections_to_delete[repo].append(collection_version.pk)
    for repo, collections in repos_with_collections_to_delete.items():
        add_and_remove(repo.pk, add_content_units=[], remove_content_units=collections)
        _cleanup_old_versions(repo)


def delete_collection_version(collection_version_pk):
    """Task to delete CollectionVersion object.

    Sequentially do the following in a single task:
    1. Call _remove_collection_version_from_repos
    2. Run orphan_cleanup to delete the CollectionVersion
    3. Delete Collection if it has no more CollectionVersion
    """
    collection_version = CollectionVersion.objects.get(pk=collection_version_pk)
    collection = collection_version.collection

    _remove_collection_version_from_repos([collection_version])

    log.info("Running orphan_cleanup to delete CollectionVersion object and artifact")
    # Running orphan_protection_time=0 should be safe since we're specifying the content
    # to be deleted. This will prevent orphan_cleanup from deleting content that is in
    # the process of being uploaded.
    orphan_cleanup(content_pks=[collection_version.pk], orphan_protection_time=0)

    if not collection.versions.exists():
        log.info("Collection has no more versions, deleting collection {}".format(collection))
        collection.delete()


def delete_collection(collection_pk):
    """Task to delete Collection object.

    Sequentially do the following in a single task:
    1. For each CollectionVersion call _remove_collection_version_from_repos
    2. Run orphan_cleanup to delete the CollectionVersions
    3. Delete Collection
    """
    collection = Collection.objects.get(pk=collection_pk)
    versions = collection.versions.all()
    _remove_collection_version_from_repos(versions)
    version_pks = versions.values_list("pk", flat=True)

    log.info("Running orphan_cleanup to delete CollectionVersion objects and artifacts")
    # Running orphan_protection_time=0 should be safe since we're specifying the content
    # to be deleted. This will prevent orphan_cleanup from deleting content that is in
    # the process of being uploaded.
    orphan_cleanup(content_pks=version_pks, orphan_protection_time=0)

    log.info("Deleting collection {}".format(collection))
    collection.delete()



from django.db import transaction
from django.db.models import Q
from pulpcore.plugin.models import RepositoryVersion

from pulp_ansible.app.models import (
    CollectionVersion,
    CollectionVersionSignature,
    AnsibleCollectionDeprecated,
    CollectionVersionMark,
    AnsibleNamespaceMetadata,
    AnsibleRepository,
)

from pulpcore.plugin.tasking import add_and_remove

from .signature import sign


@transaction.atomic
def copy_content(config):
    """
    Copy content from one repo to another in a single transaction.

    This should be an iterable of dictionaries, with each dictionary containing the following keys:

        * source_repo_version: The RepositoryVersion pk
        * dest_repo: The pk of the `AnsibleRepository` pk to copy content into. This repository will
            receive the new RepositoryVersion with the copied content.
        * `content`: An optional key that if specified should contain a list of pks to copy. These
            specified keys must be members of the `source_repo_version`. If unspecified all content
            from the `source_repo_version` will be copied to the `dest_repo`.

    Args:
        config (iterable): The config that identifies the `source_repo_version`, `dest_repo`, and
            and `content`.

    """

    def process_entry(entry):
        source_repo_version = RepositoryVersion.objects.get(pk=entry["source_repo_version"])
        dest_repo = AnsibleRepository.objects.get(pk=entry["dest_repo"])

        dest_version_provided = bool(entry.get("dest_base_version"))
        if dest_version_provided:
            dest_repo_version = RepositoryVersion.objects.get(pk=entry["dest_base_version"])
        else:
            dest_repo_version = dest_repo.latest_version()

        if entry.get("content") is not None:
            content_filter = Q(pk__in=entry.get("content"))
        else:
            content_filter = Q()

        return (
            source_repo_version,
            dest_repo_version,
            dest_repo,
            content_filter,
            dest_version_provided,
        )

    for entry in config:
        (
            source_repo_version,
            dest_repo_version,
            dest_repo,
            content_filter,
            dest_version_provided,
        ) = process_entry(entry)

        content_to_copy = source_repo_version.content.filter(content_filter)

        base_version = dest_repo_version if dest_version_provided else None
        with dest_repo.new_version(base_version=base_version) as new_version:
            new_version.add_content(content_to_copy)


def copy_collection(cv_pk_list, src_repo_pk, dest_repo_list):
    """
    Copy a list of collection versions and all of it's related contents into a
    list of destination repositories.
    """
    src_repo = AnsibleRepository.objects.get(pk=src_repo_pk)
    collection_versions = CollectionVersion.objects.filter(pk__in=cv_pk_list)

    content_types = src_repo.content.values_list("pulp_type", flat=True).distinct()
    source_pks = src_repo.content.values_list("pk", flat=True)

    content = cv_pk_list

    # collection signatures
    if "ansible.collection_signature" in content_types:
        signatures_pks = CollectionVersionSignature.objects.filter(
            signed_collection__in=cv_pk_list, pk__in=source_pks
        ).values_list("pk", flat=True)
        if signatures_pks:
            content.extend(signatures_pks)

    # collection version mark
    if "ansible.collection_mark" in content_types:
        marks_pks = CollectionVersionMark.objects.filter(
            marked_collection__in=cv_pk_list, pk__in=source_pks
        ).values_list("pk", flat=True)
        if marks_pks:
            content.extend(marks_pks)

    # namespace metadata
    namespaces = {x.namespace for x in collection_versions}
    namespaces_pks = AnsibleNamespaceMetadata.objects.filter(
        pk__in=source_pks,
        name__in=list(namespaces),
    ).values_list("pk", flat=True)
    if namespaces_pks:
        content.extend(namespaces_pks)

    # collection deprecation
    for cv in collection_versions:
        if "ansible.collection_deprecation" in content_types:
            deprecations_pks = AnsibleCollectionDeprecated.objects.filter(
                pk__in=source_pks,
                namespace=cv.namespace,
                name=cv.name,
            ).values_list("pk", flat=True)
            if deprecations_pks:
                content.extend(deprecations_pks)

    for pk in dest_repo_list:
        add_and_remove(repository_pk=pk, add_content_units=content, remove_content_units=[])


def move_collection(cv_pk_list, src_repo_pk, dest_repo_list):
    """
    Copy a collection and all of it's contents into a list of destination repositories
    """
    copy_collection(cv_pk_list, src_repo_pk, dest_repo_list)

    # No need to remove anything other than the collection version because everything
    # else will get handled by AnsibleRepository.finalize_repo_version()
    add_and_remove(repository_pk=src_repo_pk, add_content_units=[], remove_content_units=cv_pk_list)


def copy_or_move_and_sign(copy_or_move, signing_service_pk=None, **move_task_kwargs):
    if copy_or_move == "copy":
        copy_collection(**move_task_kwargs)
    else:
        move_collection(**move_task_kwargs)

    if signing_service_pk:
        for pk in move_task_kwargs["dest_repo_list"]:
            sign(
                signing_service_href=signing_service_pk,
                repository_href=pk,
                content_hrefs=move_task_kwargs["cv_pk_list"],
            )



# Generated by Django 4.2.22 on 2025-07-01 14:04

from django.db import migrations, models
import django.db.models.deletion
import pulpcore.app.util


class Migration(migrations.Migration):

    dependencies = [
        ("core", "0123_upstreampulp_q_select"),
        ("ansible", "0063_domain_support"),
    ]

    operations = [
        migrations.AlterUniqueTogether(
            name="collection",
            unique_together=set(),
        ),
        migrations.AlterUniqueTogether(
            name="collectiondownloadcount",
            unique_together=set(),
        ),
        migrations.AddField(
            model_name="ansiblenamespace",
            name="pulp_domain",
            field=models.ForeignKey(
                default=pulpcore.app.util.get_domain_pk,
                on_delete=django.db.models.deletion.PROTECT,
                to="core.domain",
            ),
        ),
        migrations.AddField(
            model_name="collection",
            name="pulp_domain",
            field=models.ForeignKey(
                default=pulpcore.app.util.get_domain_pk,
                on_delete=django.db.models.deletion.PROTECT,
                to="core.domain",
            ),
        ),
        migrations.AddField(
            model_name="collectiondownloadcount",
            name="pulp_domain",
            field=models.ForeignKey(
                default=pulpcore.app.util.get_domain_pk,
                on_delete=django.db.models.deletion.PROTECT,
                to="core.domain",
            ),
        ),
        migrations.AlterField(
            model_name="ansiblenamespace",
            name="name",
            field=models.CharField(max_length=64),
        ),
        migrations.AlterUniqueTogether(
            name="ansiblenamespace",
            unique_together={("pulp_domain", "name")},
        ),
        migrations.AlterUniqueTogether(
            name="collection",
            unique_together={("pulp_domain", "namespace", "name")},
        ),
        migrations.AlterUniqueTogether(
            name="collectiondownloadcount",
            unique_together={("pulp_domain", "namespace", "name")},
        ),
    ]



"""Tests related to Galaxy V3 serializers."""


def test_v3_updated_at(
    ansible_bindings,
    ansible_sync_factory,
    ansible_collection_remote_factory,
    ansible_distribution_factory,
):
    """Test Collections V3 endpoint field: ``updated_at``."""

    requirements1 = """
    collections:
        - name: "pulp.squeezer"
          version: "0.0.7"
    """

    requirements2 = """
    collections:
        - name: "pulp.squeezer"
          version: "0.0.17"
    """

    # sync the first version ...
    remote = ansible_collection_remote_factory(
        url="https://galaxy.ansible.com",
        requirements_file=requirements1,
        sync_dependencies=False,
    )

    repository = ansible_sync_factory(remote=remote.pulp_href)
    distribution = ansible_distribution_factory(repository=repository)

    collections = ansible_bindings.PulpAnsibleApiV3CollectionsApi.list(distribution.base_path)

    original_highest_version = collections.data[0].highest_version["version"]
    original_updated_at = collections.data[0].updated_at

    original_total_versions = ansible_bindings.PulpAnsibleApiV3CollectionsVersionsApi.list(
        "squeezer", "pulp", distribution.base_path
    ).meta.count

    # sync the second version ...
    # Don't wait on the update task here. The sync will wait for us.
    ansible_bindings.RemotesCollectionApi.partial_update(
        remote.pulp_href, {"requirements_file": requirements2}
    )

    ansible_sync_factory(ansible_repo=repository, remote=remote.pulp_href, optimize=True)

    # enumerate new data after 2nd sync ...
    collections = ansible_bindings.PulpAnsibleApiV3CollectionsApi.list(distribution.base_path)
    highest_version = collections.data[0].highest_version["version"]
    updated_at = collections.data[0].updated_at
    total_versions = ansible_bindings.PulpAnsibleApiV3CollectionsVersionsApi.list(
        "squeezer", "pulp", distribution.base_path
    ).meta.count

    assert original_highest_version == "0.0.7"
    assert highest_version == "0.0.17"
    assert original_total_versions == 1
    assert total_versions == 2
    assert updated_at > original_updated_at


def test_v3_collection_version_from_synced_data(
    ansible_bindings,
    ansible_sync_factory,
    ansible_collection_remote_factory,
    ansible_distribution_factory,
):
    """Test Collection Versions V3 endpoint fields."""
    remote = ansible_collection_remote_factory(
        url="https://galaxy.ansible.com",
        requirements_file="collections:\n  - name: cisco.nxos\n    version: 1.4.0",
        sync_dependencies=False,
    )

    repository = ansible_sync_factory(remote=remote.pulp_href)
    distribution = ansible_distribution_factory(repository=repository)

    version = ansible_bindings.PulpAnsibleApiV3CollectionsVersionsApi.read(
        "nxos", "cisco", distribution.base_path, "1.4.0"
    )

    assert version.requires_ansible == ">=2.9.10,<2.11"
    assert "'name': 'README.md'" in str(version.files)
    assert version.manifest["collection_info"]["name"] == "nxos"



import logging
from semantic_version import Version

from django.db.models import Q

from pulp_ansible.app.models import (
    AnsibleDistribution,
    AnsibleCollectionDeprecated,
    AnsibleNamespaceMetadata,
    AnsibleRepository,
    CollectionVersion,
    CollectionVersionSignature,
    CrossRepositoryCollectionVersionIndex as CVIndex,
)

from pulpcore.plugin.models import RepositoryVersion


log = logging.getLogger(__name__)


def has_distribution(repository, repository_version):
    """Is -any- distribution attached to this repo or repo version?"""
    return AnsibleDistribution.objects.filter(
        Q(repository=repository) | Q(repository_version=repository_version)
    ).exists()


def get_highest_version_string_from_cv_objects(cv_objects):
    """Return the highest version as a string (preferring stable releases)."""
    versions_strings = [x.version for x in cv_objects]
    versions = [Version(x) for x in versions_strings]
    versions = sorted(versions, reverse=True)
    stable_versions = [x for x in versions if not x.prerelease]
    if stable_versions:
        return str(stable_versions[0])
    return str(versions[0])


def compute_repository_changes(repository_version):
    """Use the previous version to make a list of namespace(s).name(s) changed."""
    # Figure out what the previous repo version is
    repository = repository_version.repository
    previous_number = repository_version.number - 1
    previous_version = RepositoryVersion.objects.filter(
        repository=repository, number=previous_number
    ).first()

    # If there isn't a previous verison, all things have "changed"
    if previous_version is None:
        return None

    changed_collections = set()

    cv_type = CollectionVersion.get_pulp_type()
    deprecation_type = AnsibleCollectionDeprecated.get_pulp_type()
    signature_type = CollectionVersionSignature.get_pulp_type()
    metadata_type = AnsibleNamespaceMetadata.get_pulp_type()

    for func in [repository_version.added, repository_version.removed]:
        for modified in func(base_version=previous_version):
            if modified.pulp_type == cv_type:
                cv = modified.ansible_collectionversion
                changed_collections.add((cv.namespace, cv.name))
            elif modified.pulp_type == deprecation_type:
                deprecation = modified.ansible_ansiblecollectiondeprecated
                changed_collections.add((deprecation.namespace, deprecation.name))
            elif modified.pulp_type == signature_type:
                signature = modified.ansible_collectionversionsignature
                changed_collections.add(
                    (signature.signed_collection.namespace, signature.signed_collection.name)
                )
            elif modified.pulp_type == metadata_type:
                metadata = modified.ansible_ansiblenamespacemetadata
                changed_collections.add((metadata.name, "*"))

    return changed_collections


def update_index(distribution=None, repository=None, repository_version=None, is_latest=False):
    """Rebuild index by distribtion|repository|repositoryversion."""

    # if the distro points at a specific repo version, we should use that in the index
    # otherwise the index value for repository version should be null
    # use_repository_version = False
    use_repository_version = not is_latest

    # a repov was passed in so we should use that
    # if repository_version:
    #    use_repository_version = True

    # make sure the distro points at a repo[version]
    if distribution and not repository and not repository_version:
        # sometimes distros point at a version
        if distribution.repository_version:
            repository = distribution.repository_version.repository
            repository_version = distribution.repository_version
            use_repository_version = True

        # sometimes distros point at a repository
        elif distribution.repository is not None:
            repository = distribution.repository
            repository_version = distribution.repository.latest_version()
            # is_latest = True
            use_repository_version = False

        # sometimes distros point at nothing?
        else:
            return

    # extract repository from repository version if needed
    if repository is None:
        repository = repository_version.repository

    # optimization: -must- have an AnsibleRepository for the index model
    if not isinstance(repository, AnsibleRepository):
        repository = AnsibleRepository.objects.filter(pk=repository.pk).first()
        if repository is None:
            return

    # optimization: we only want to index "distributed" CVs
    if distribution is None and not has_distribution(repository, repository_version):
        return

    # This block handles a case where the distribution that points at a repository
    # has been deleted. If no other distribution points at the repository, all related
    # indexes need to be removed and to exit early.
    if not use_repository_version:
        if not has_distribution(repository, repository_version):
            CVIndex.objects.filter(repository=repository, repository_version=None).delete()
            return

    # optimizaion: exit early if using a repo version and it's alreay indexed
    if use_repository_version:
        if CVIndex.objects.filter(repository_version=repository_version).exists():
            return

    # What has changed between this version and the last?
    changed_collections = compute_repository_changes(repository_version)
    if not changed_collections:
        return

    # get all CVs in this repository version
    cvs_pks = repository_version.content.filter(pulp_type="ansible.collection_version").values_list(
        "pk", flat=True
    )
    cvs = CollectionVersion.objects.filter(pk__in=cvs_pks)

    # get the set of signatures in this repo version
    repo_signatures_pks = repository_version.content.filter(
        pulp_type="ansible.collection_signature"
    ).values_list("pk", flat=True)
    repo_signatures = CollectionVersionSignature.objects.filter(pk__in=repo_signatures_pks)

    # get the set of deprecations in this repo version
    deprecations = repository_version.content.filter(
        pulp_type="ansible.collection_deprecation"
    ).values_list("pk", flat=True)
    deprecations = AnsibleCollectionDeprecated.objects.filter(pk__in=deprecations)
    deprecations_set = {(x.namespace, x.name) for x in deprecations}

    # find all the most recent namespace metadata in the repo version
    namespaces = {}
    for ns in repository_version.get_content(content_qs=AnsibleNamespaceMetadata.objects).all():
        if ns.name not in namespaces:
            namespaces[ns.name] = ns
            continue
        if namespaces[ns.name].timestamp_of_interest < ns.timestamp_of_interest:
            namespaces[ns.name] = ns

    # map out the namespace(s).name(s) for everything in the repo version
    colset = set(cvs.values_list("namespace", "name").distinct())

    repo_v = None
    if use_repository_version:
        repo_v = repository_version

    # clean out cvs no longer in the repo when a distro w/ a repo
    if not use_repository_version:
        CVIndex.objects.filter(repository=repository, repository_version=None).exclude(
            collection_version__pk__in=cvs
        ).delete()

    # iterate through each collection in the repository
    for colkey in colset:
        namespace, name = colkey

        if (namespace, name) not in changed_collections and (
            namespace,
            "*",
        ) not in changed_collections:
            continue

        # get all the versions for this collection
        related_cvs = cvs.filter(namespace=namespace, name=name).only("version")

        # what is the "highest" version in this list?
        highest_version = get_highest_version_string_from_cv_objects(related_cvs)

        # should all of these CVs be deprecated?
        is_deprecated = colkey in deprecations_set

        # process each related CV
        for rcv in related_cvs:
            # get the related signatures for this CV
            rcv_signatures = repo_signatures.filter(signed_collection=rcv).count()

            # create|update the index for this CV
            CVIndex.objects.update_or_create(
                repository=repository,
                repository_version=repo_v,
                collection_version=rcv,
                defaults={
                    "is_highest": rcv.version == highest_version,
                    "is_signed": rcv_signatures > 0,
                    "is_deprecated": is_deprecated,
                    "namespace_metadata": namespaces.get(namespace, None),
                },
            )


def update_distribution_index(distribution):
    return update_index(distribution=distribution)


def rebuild_index():
    """Rebuild -everything-."""
    indexed_repos = set()
    dqs = AnsibleDistribution.objects.select_related(
        "repository", "repository_version", "repository_version__repository"
    ).all()
    for distro in dqs:
        if distro.repository_version:
            rv = distro.repository_version
        else:
            rv = distro.repository.latest_version()

        if rv.pk in indexed_repos:
            continue

        update_index(distribution=distro)
        indexed_repos.add(rv.pk)



# Generated by Django 4.2.19 on 2025-05-21 13:50

import django.contrib.postgres.fields
from django.db import migrations, models
from pulpcore.plugin.migrations import RequireVersion


def migrate_missing_tags_up(apps, schema):
    CollectionVersion = apps.get_model("ansible", "CollectionVersion")
    cv_to_update = []
    for cv in CollectionVersion.objects.filter(new_tags=None).only("new_tags").prefetch_related("tags"):
        cv.new_tags = [t.name for t in cv.tags.all()]
        cv_to_update.append(cv)
        if len(cv_to_update) >= 100:
            CollectionVersion.objects.bulk_update(cv_to_update, ["new_tags"])
            cv_to_update = []
    if len(cv_to_update) > 0:
        CollectionVersion.objects.bulk_update(cv_to_update, ["new_tags"])


def migrate_missing_tags_down(apps, schema):
    # Warning, this Down-Migration is not Zero-Downtime safe.
    # Also it's not optimized.
    Tag = apps.get_model("ansible", "Tag")
    CollectionVersion = apps.get_model("ansible", "CollectionVersion")
    for cv in CollectionVersion.objects.filter(new_tags__len__gt=0).only("new_tags"):
        for tag_name in cv.new_tags:
            tag, created = Tag.objects.get_or_create(name=tag_name)
            cv.tags.add(tag)


# This query generates full text search index based
# the following data ranked from A to D:
#   - Namespace name (weight A)
#   - Collection name (weight A)
#   - Collection tags (weight B)
#   - Collection content names (weight C)
#   - Collection description (weight D)
NEW_TS_VECTOR_SELECT = """
    setweight(to_tsvector(coalesce(NEW.namespace,'')), 'A')
    || setweight(to_tsvector(coalesce(NEW.name, '')), 'A')
    || setweight(to_tsvector(array_to_string(coalesce(NEW.tags, ARRAY[]::bpchar[]), ' ')), 'B')
    || (
      SELECT
        setweight(to_tsvector(coalesce(string_agg(cvc ->> 'name', ' '), '')), 'C')
      FROM jsonb_array_elements(NEW.contents) AS cvc
    )
    || setweight(to_tsvector(coalesce(NEW.description, '')), 'D')
"""

OLD_TS_VECTOR_SELECT = """
    SELECT
    setweight(to_tsvector(coalesce(namespace,'')), 'A')
    || setweight(to_tsvector(coalesce(name, '')), 'A')
    || (
      SELECT
        setweight(to_tsvector(
          coalesce(string_agg("ansible_tag"."name", ' '), '')
        ), 'B')
      FROM
        "ansible_tag" INNER JOIN "ansible_collectionversion_tags"
        ON
        ("ansible_tag"."pulp_id" = "ansible_collectionversion_tags"."tag_id" )
      WHERE
        "ansible_collectionversion_tags"."collectionversion_id" = cv.content_ptr_id
    )
    || (
      SELECT
        setweight(to_tsvector(
          coalesce(string_agg(cvc ->> 'name', ' '), '')
        ), 'C')
      FROM jsonb_array_elements(cv.contents) AS cvc
    )
    || setweight(to_tsvector(coalesce(description, '')), 'D')
    FROM ansible_collectionversion cv
    WHERE cv.content_ptr_id = NEW.content_ptr_id
"""

# Creates a database function and a trigger to update collection search
# vector field.
# In the new version, tags are availabel at create and content should never be updated,
# so having both insert and update in the trigger from now on is probably fast _and_ safe.
UPDATE_COLLECTIONS_TS_VECTOR_TRIGGER = """
SET CONSTRAINTS ALL IMMEDIATE;
CREATE OR REPLACE FUNCTION update_collection_ts_vector()
    RETURNS TRIGGER AS
$$
BEGIN
    NEW.search_vector := ({TS_VECTOR_SELECT});
    RETURN NEW;
END;
$$ LANGUAGE plpgsql;
DROP TRIGGER IF EXISTS update_ts_vector ON ansible_collectionversion;
CREATE TRIGGER update_ts_vector
    {WHEN}
    ON ansible_collectionversion
    FOR EACH ROW
EXECUTE FUNCTION update_collection_ts_vector();
"""


class Migration(migrations.Migration):

    dependencies = [
        ("ansible", "0061_collectionversion_new_tags"),
    ]

    operations = [
        RequireVersion("ansible", "0.26.0"),
        migrations.RunPython(code=migrate_missing_tags_up, reverse_code=migrate_missing_tags_down),
        migrations.RunSQL(
            sql=UPDATE_COLLECTIONS_TS_VECTOR_TRIGGER.format(
                TS_VECTOR_SELECT=NEW_TS_VECTOR_SELECT,
                WHEN="BEFORE INSERT OR UPDATE",
            ),
            reverse_sql=UPDATE_COLLECTIONS_TS_VECTOR_TRIGGER.format(
                TS_VECTOR_SELECT=OLD_TS_VECTOR_SELECT,
                WHEN="BEFORE UPDATE",
            ),
        ),
        migrations.RemoveField(
            model_name="collectionversion",
            name="tags",
        ),
        migrations.DeleteModel(
            name="Tag",
        ),
        migrations.RenameField(
            model_name="collectionversion",
            old_name="new_tags",
            new_name="tags",
        ),
        migrations.AlterField(
            model_name="collectionversion",
            name="tags",
            field=django.contrib.postgres.fields.ArrayField(
                base_field=models.CharField(max_length=64, unique=True), default=list, size=None
            ),
        ),
    ]



import re

from django.shortcuts import get_object_or_404
from drf_spectacular.utils import extend_schema, extend_schema_view
from rest_framework import generics, response, views
from pulpcore.plugin.util import get_domain

from pulp_ansible.app.models import AnsibleDistribution, Role

from .serializers import GalaxyRoleSerializer, GalaxyRoleVersionSerializer


class DistributionMixin:
    """
    Mixin for getting the content for the view.
    """

    @property
    def _repository_version(self):
        """Returns repository version."""
        path = self.kwargs["path"]

        context = getattr(self, "pulp_context", None)
        if context and context.get(path, None):
            return self.pulp_context[path]

        distro = get_object_or_404(AnsibleDistribution, base_path=path)
        if distro.repository_version:
            self.pulp_context = {path: distro.repository_version}
            return distro.repository_version

        repo_version = distro.repository.latest_version() if distro.repository else None
        self.pulp_context = {path: repo_version}
        return repo_version

    @property
    def _distro_content(self):
        """Returns distribution content."""
        repo_version = self._repository_version
        if repo_version is None:
            return self.model.objects.none()

        return repo_version.content

    def get_serializer_context(self):
        """Inserts distribution path to a serializer context."""
        context = super().get_serializer_context()
        if "path" in self.kwargs:
            context["path"] = self.kwargs["path"]
        return context


class GalaxyVersionView(views.APIView):
    """
    APIView for Galaxy versions.
    """

    authentication_classes = []
    permission_classes = []

    v3_only = False

    def get(self, request, **kwargs):
        """
        Return a response to the "GET" action.
        """
        available_versions = {"v1": "v1/", "v3": "v3/"}
        if self.v3_only:
            available_versions = {"v3": "v3/"}

        api_info = {
            "available_versions": available_versions,
            "current_version": "v3",
        }

        return response.Response(api_info)


class RoleList(DistributionMixin, generics.ListAPIView):
    """
    APIView for Roles.
    """

    model = Role
    serializer_class = GalaxyRoleSerializer
    authentication_classes = []
    permission_classes = []

    def get_queryset(self):
        """
        Get the list of items for this view.
        """
        roles = Role.objects.distinct("namespace", "name").filter(
            pk__in=self._distro_content, pulp_domain=get_domain()
        )

        namespace = self.request.query_params.get("owner__username", None)
        if namespace:
            roles = roles.filter(namespace__iexact=namespace)
        name = self.request.query_params.get("name", None)
        if name:
            roles = roles.filter(name__iexact=name)

        return roles


@extend_schema_view(get=extend_schema(operation_id="api_v1_roles_versions_list"))
class RoleVersionList(DistributionMixin, generics.ListAPIView):
    """
    APIView for Role Versions.
    """

    model = Role
    serializer_class = GalaxyRoleVersionSerializer
    authentication_classes = []
    permission_classes = []

    def get_queryset(self):
        """
        Get the list of items for this view.
        """
        namespace, name = re.split(r"\.", self.kwargs["role_pk"])
        versions = Role.objects.filter(
            pk__in=self._distro_content, name=name, namespace=namespace, pulp_domain=get_domain()
        )
        return versions



# Generated by Django 4.2.4 on 2023-09-15 09:23

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("ansible", "0054_split_collection_version_numbers"),
    ]

    operations = [
        migrations.RunSQL(
            # Custom roles are only allowed with postgreSQL 16.
            sql="CREATE COLLATION pulp_ansible_semver (provider = 'icu', locale = 'en-US-u-kn-true')",
            reverse_sql="DROP COLLATION pulp_ansible_semver",
        ),
        migrations.AlterField(
            model_name="collectionversion",
            name="version",
            field=models.CharField(db_collation="pulp_ansible_semver", max_length=128),
        ),
        migrations.AlterField(
            model_name="role",
            name="version",
            field=models.CharField(db_collation="pulp_ansible_semver", max_length=128),
        ),
    ]



import argparse

import django

django.setup()

from pulpcore.plugin.tasking import (  # noqa otherwise E402: module level not at top of file
    dispatch,
)

from pulp_ansible.app.tasks.test_tasks import (  # noqa otherwise E402: module level not at top of file
    create_repos_with_collections,
)


parser = argparse.ArgumentParser(
    description="Create Repositories with Collections and several " "RepositoryVersions."
)

parser.add_argument(
    "--num-repos",
    metavar="N",
    type=int,
    nargs=1,
    default=[50000],
    help="The number of Ansible Repositories to create.",
)

parser.add_argument(
    "--num-repo-versions-per-repo",
    metavar="N",
    type=int,
    nargs=1,
    default=[10],
    help="The number of RepositoryVersions to create per Repository.",
)

parser.add_argument(
    "--batch-size",
    metavar="N",
    type=int,
    nargs=1,
    default=[1000],
    help="The number of Ansible Repositories to create per batch.",
)

parser.add_argument(
    "--collection-percentage",
    metavar="PERCENT",
    type=float,
    nargs=1,
    default=[0.9],
    help="The percentage of collections each Repository should "
    "contain. For example 0.9 would have 90% of all collections"
    "in Pulp be associated with each repository. Collections "
    "are selected at random.",
)

args = parser.parse_args()


if __name__ == "__main__":
    repos_to_still_make = args.num_repos[0]
    num_repo_versions_per_repo = args.num_repo_versions_per_repo[0]
    collection_percentage = args.collection_percentage[0]

    while repos_to_still_make > 0:
        if repos_to_still_make >= args.batch_size[0]:
            task_args = (args.batch_size[0], num_repo_versions_per_repo, collection_percentage)
            async_result = dispatch(create_repos_with_collections, args=task_args)
            print("Dispatched task with {} repositories".format(args.batch_size[0]))
            repos_to_still_make = repos_to_still_make - args.batch_size[0]
        else:
            task_args = (repos_to_still_make, num_repo_versions_per_repo, collection_percentage)
            async_result = dispatch(create_repos_with_collections, args=task_args)
            print("Dispatched task with {} repositories".format(repos_to_still_make))
            repos_to_still_make = 0



import hashlib
import json
from logging import getLogger

from semantic_version import Version

from django.conf import settings
from django.db import models
from django.db.utils import IntegrityError
from django.contrib.postgres import fields as psql_fields
from django.contrib.postgres import search as psql_search
from django_lifecycle import (
    AFTER_UPDATE,
    AFTER_DELETE,
    AFTER_CREATE,
    BEFORE_UPDATE,
    BEFORE_SAVE,
    hook,
)

from pulpcore.plugin.models import (
    AutoAddObjPermsMixin,
    BaseModel,
    Content,
    Remote,
    Repository,
    RepositoryVersion,
    Distribution,
    SigningService,
    Task,
    EncryptedTextField,
)
from pulpcore.plugin.repo_version_utils import remove_duplicates, validate_repo_version
from pulpcore.plugin.util import get_domain_pk

from .downloaders import AnsibleDownloaderFactory


log = getLogger(__name__)


class Role(Content):
    """
    A content type representing a Role.
    """

    TYPE = "role"

    namespace = models.CharField(max_length=64)
    name = models.CharField(max_length=64)
    version = models.CharField(max_length=128, db_collation="pulp_ansible_semver")
    _pulp_domain = models.ForeignKey("core.Domain", default=get_domain_pk, on_delete=models.PROTECT)

    @property
    def relative_path(self):
        """
        Return the relative path of the ContentArtifact.
        """
        return self.contentartifact_set.get().relative_path

    class Meta:
        default_related_name = "%(app_label)s_%(model_name)s"
        unique_together = ("version", "name", "namespace", "_pulp_domain")


class Collection(BaseModel):
    """A model representing a Collection."""

    namespace = models.CharField(max_length=64, editable=False)
    name = models.CharField(max_length=64, editable=False)
    pulp_domain = models.ForeignKey("core.Domain", default=get_domain_pk, on_delete=models.PROTECT)

    def __str__(self):
        """Return a representation."""
        return f"<{self.__class__.__name__}: {self.namespace}.{self.name}>"

    class Meta:
        unique_together = ("pulp_domain", "namespace", "name")


class CollectionImport(models.Model):
    """A model representing a collection import task details."""

    task = models.OneToOneField(
        Task, on_delete=models.CASCADE, editable=False, related_name="+", primary_key=True
    )
    messages = models.JSONField(default=list, editable=False)

    class Meta:
        ordering = ["task__pulp_created"]

    def add_log_record(self, log_record):
        """
        Records a single log message but does not save the CollectionImport object.

        Args:
            log_record(logging.LogRecord): The logging record to record on messages.

        """
        self.messages.append(
            {"message": log_record.msg, "level": log_record.levelname, "time": log_record.created}
        )


class AnsibleNamespace(BaseModel):
    """
    A model representing a Namespace. This should be used for permissions.
    """

    name = models.CharField(max_length=64, editable=True)
    pulp_domain = models.ForeignKey("core.Domain", default=get_domain_pk, on_delete=models.PROTECT)

    class Meta:
        unique_together = ("pulp_domain", "name")


class CollectionVersion(Content):
    """
    A content type representing a CollectionVersion.

    This model is primarily designed to adhere to the data format for Collection content. That spec
    is here: https://docs.ansible.com/ansible/devel/dev_guide/collections_galaxy_meta.html

    Fields:

        authors (psql_fields.ArrayField): A list of the CollectionVersion content's authors.
        contents (models.JSONField): A JSON field with data about the contents.
        dependencies (models.JSONField): A dict declaring Collections that this collection
            requires to be installed for it to be usable.
        description (models.TextField): A short summary description of the collection.
        docs_blob (models.JSONField): A JSON field holding the various documentation blobs in
            the collection.
        manifest (models.JSONField): A JSON field holding MANIFEST.json data.
        files (models.JSONField): A JSON field holding FILES.json data.
        documentation (models.CharField): The URL to any online docs.
        homepage (models.CharField): The URL to the homepage of the collection/project.
        issues (models.CharField): The URL to the collection issue tracker.
        license (psql_fields.ArrayField): A list of licenses for content inside of a collection.
        name (models.CharField): The name of the collection.
        namespace (models.CharField): The namespace of the collection.
        repository (models.CharField): The URL of the originating SCM repository.
        version (models.CharField): The version of the collection.
        requires_ansible (models.CharField): The version of Ansible required to use the collection.
        tag (ArrayField): Taglist.

    Relations:

        collection (models.ForeignKey): Reference to a collection model.
    """

    TYPE = "collection_version"
    repo_key_fields = ("name", "namespace", "version")

    # Data Fields
    authors = psql_fields.ArrayField(models.CharField(max_length=64), default=list, editable=False)
    contents = models.JSONField(default=list, editable=False)
    dependencies = models.JSONField(default=dict, editable=False)
    description = models.TextField(default="", blank=True, editable=False)
    docs_blob = models.JSONField(default=dict, editable=False)
    manifest = models.JSONField(default=dict, editable=False)
    files = models.JSONField(default=dict, editable=False)
    documentation = models.CharField(default="", blank=True, max_length=2000, editable=False)
    homepage = models.CharField(default="", blank=True, max_length=2000, editable=False)
    issues = models.CharField(default="", blank=True, max_length=2000, editable=False)
    license = psql_fields.ArrayField(models.CharField(max_length=32), default=list, editable=False)
    name = models.CharField(max_length=64, editable=False)
    namespace = models.CharField(max_length=64, editable=False)
    repository = models.CharField(default="", blank=True, max_length=2000, editable=False)
    requires_ansible = models.CharField(null=True, max_length=255)
    sha256 = models.CharField(max_length=64, db_index=True, null=False, blank=False)

    version = models.CharField(max_length=128, db_collation="pulp_ansible_semver")
    version_major = models.IntegerField()
    version_minor = models.IntegerField()
    version_patch = models.IntegerField()
    version_prerelease = models.CharField(max_length=128)
    tags = psql_fields.ArrayField(models.CharField(max_length=64, unique=True), default=list)

    # Foreign Key Fields
    collection = models.ForeignKey(
        Collection, on_delete=models.PROTECT, related_name="versions", editable=False
    )

    # Search Fields
    #   This field is populated by a trigger setup in the database by
    #   a migration file. The trigger only runs when the table is
    #   updated. CollectionVersions are INSERT'ed into the table, so
    #   the search_vector does not get populated at initial creation
    #   time.
    search_vector = psql_search.SearchVectorField(default="")
    _pulp_domain = models.ForeignKey("core.Domain", default=get_domain_pk, on_delete=models.PROTECT)

    @hook(BEFORE_SAVE)
    def calculate_version_parts(self):
        v = Version(self.version)

        self.version_major = v.major
        self.version_minor = v.minor
        self.version_patch = v.patch
        self.version_prerelease = ".".join(v.prerelease)

    @property
    def relative_path(self):
        """
        Return the relative path for the ContentArtifact.
        """
        return "{namespace}-{name}-{version}.tar.gz".format(
            namespace=self.namespace, name=self.name, version=self.version
        )

    def __str__(self):
        """Return a representation."""
        return f"<{self.__class__.__name__}: {self.namespace}.{self.name} {self.version}>"

    class Meta:
        default_related_name = "%(app_label)s_%(model_name)s"
        unique_together = ("sha256", "_pulp_domain")
        indexes = [
            models.Index(fields=["namespace", "name", "version"]),
        ]


class CollectionVersionMark(Content):
    """
    A content type representing a mark that is attached to a content unit.

    Fields:
        value (models.CharField): The value of the mark.
        marked_collection (models.ForeignKey): Reference to a CollectionVersion.
    """

    PROTECTED_FROM_RECLAIM = False
    TYPE = "collection_mark"

    value = models.SlugField()
    marked_collection = models.ForeignKey(
        CollectionVersion, null=False, on_delete=models.CASCADE, related_name="marks"
    )
    _pulp_domain = models.ForeignKey("core.Domain", default=get_domain_pk, on_delete=models.PROTECT)

    class Meta:
        default_related_name = "%(app_label)s_%(model_name)s"
        unique_together = ("value", "marked_collection", "_pulp_domain")


class CollectionVersionSignature(Content):
    """
    A content type representing a signature that is attached to a content unit.

    Fields:
        data (models.BinaryField): A signature, base64 encoded. # Not sure if it is base64 encoded
        digest (models.CharField): A signature sha256 digest.
        pubkey_fingerprint (models.CharField): A fingerprint of the public key used.

    Relations:
        signed_collection (models.ForeignKey): A collection version this signature is relevant to.
        signing_service (models.ForeignKey): An optional signing service used for creation.
    """

    PROTECTED_FROM_RECLAIM = False
    TYPE = "collection_signature"

    signed_collection = models.ForeignKey(
        CollectionVersion, on_delete=models.CASCADE, related_name="signatures"
    )
    data = models.TextField()
    digest = models.CharField(max_length=64)
    pubkey_fingerprint = models.CharField(max_length=64)
    signing_service = models.ForeignKey(
        SigningService, on_delete=models.SET_NULL, related_name="signatures", null=True
    )
    _pulp_domain = models.ForeignKey("core.Domain", default=get_domain_pk, on_delete=models.PROTECT)

    class Meta:
        default_related_name = "%(app_label)s_%(model_name)s"
        unique_together = ("pubkey_fingerprint", "signed_collection", "_pulp_domain")


class AnsibleNamespaceMetadata(Content):
    """
    A content type representing the Namespace metadata of a Collection.

    Namespace metadata can have an avatar which is stored as an associated artifact upon upload.

    Fields:
        name (models.CharField): The required name of the Namespace
        company (models.CharField): Optional namespace owner company name.
        email (models.CharField): Optional namespace contact email.
        description (models.CharField): Namespace brief description.
        resources (models.TextField): Namespace resources page in markdown format.
        links (psql_fields.HStore): Labeled related links.
        avatar_sha256 (models.CharField): SHA256 digest of avatar image.
        metadata_sha256(models.CharField): SHA256 digest of all other metadata fields.

    Relations:
        namespace (AnsibleNamespace): A link to the Namespace model used for permissions.
    """

    TYPE = "namespace"
    repo_key_fields = ("name",)
    # fields on the existing model in galaxy_ng
    name = models.CharField(max_length=64, blank=False)
    company = models.CharField(max_length=64, blank=True, default="")
    email = models.CharField(max_length=256, blank=True, default="")
    description = models.CharField(max_length=256, blank=True, default="")
    resources = models.TextField(blank=True, default="")

    links = psql_fields.HStoreField(default=dict)
    avatar_sha256 = models.CharField(max_length=64, null=True)

    # Hash of the values of all the fields mentioned above.
    # Content uniqueness constraint.
    metadata_sha256 = models.CharField(max_length=64, db_index=True, blank=False)
    namespace = models.ForeignKey(
        AnsibleNamespace, on_delete=models.PROTECT, related_name="metadatas"
    )
    _pulp_domain = models.ForeignKey("core.Domain", default=get_domain_pk, on_delete=models.PROTECT)

    @property
    def avatar_artifact(self):
        return self._artifacts.filter(
            pulp_domain=get_domain_pk(), content_memberships__relative_path=f"{self.name}-avatar"
        ).first()

    @hook(BEFORE_SAVE)
    def calculate_metadata_sha256(self):
        """Calculates the metadata_sha256 from the other metadata fields."""
        metadata = {
            "name": self.name,
            "company": self.company,
            "email": self.email,
            "description": self.description,
            "resources": self.resources,
            "links": self.links,
            "avatar_sha256": self.avatar_sha256,
        }

        metadata_json = json.dumps(metadata, sort_keys=True).encode("utf-8")
        hasher = hashlib.sha256(metadata_json)
        if self.metadata_sha256:
            # If we are creating from sync, assert that calculated hash == synced hash
            if self.metadata_sha256 != hasher.hexdigest():
                raise IntegrityError("Calculated digest does not equal passed in digest")
        else:
            self.metadata_sha256 = hasher.hexdigest()

    class Meta:
        default_related_name = "%(app_label)s_%(model_name)s"
        unique_together = ("metadata_sha256", "_pulp_domain")


class DownloadLog(BaseModel):
    """
    A download log for content units by user, IP and org_id.
    """

    content_unit = models.ForeignKey(
        Content, on_delete=models.CASCADE, related_name="download_logs"
    )
    user = models.ForeignKey(
        settings.AUTH_USER_MODEL,
        on_delete=models.CASCADE,
        null=True,
        related_name="download_logs",
    )
    ip = models.GenericIPAddressField()
    extra_data = models.JSONField(null=True)
    user_agent = models.TextField()
    repository = models.ForeignKey(
        Repository, on_delete=models.CASCADE, related_name="download_logs"
    )
    repository_version = models.ForeignKey(
        RepositoryVersion, null=True, on_delete=models.SET_NULL, related_name="download_logs"
    )


class CollectionDownloadCount(BaseModel):
    """
    Aggregate count of downloads per collection
    """

    pulp_domain = models.ForeignKey("core.Domain", default=get_domain_pk, on_delete=models.PROTECT)
    namespace = models.CharField(max_length=64, editable=False, db_index=True)
    name = models.CharField(max_length=64, editable=False, db_index=True)
    download_count = models.BigIntegerField(default=0)

    class Meta:
        unique_together = ("pulp_domain", "namespace", "name")


class RoleRemote(Remote, AutoAddObjPermsMixin):
    """
    A Remote for Ansible content.
    """

    TYPE = "role"

    class Meta:
        default_related_name = "%(app_label)s_%(model_name)s"
        permissions = [("manage_roles_roleremote", "Can manage roles on role remotes")]


class CollectionRemote(Remote, AutoAddObjPermsMixin):
    """
    A Remote for Collection content.
    """

    TYPE = "collection"

    requirements_file = models.TextField(null=True)
    auth_url = models.CharField(null=True, max_length=255)
    token = EncryptedTextField(null=True)
    sync_dependencies = models.BooleanField(default=True)
    signed_only = models.BooleanField(default=False)

    @property
    def download_factory(self):
        """
        Return the DownloaderFactory which can be used to generate asyncio capable downloaders.

        Upon first access, the DownloaderFactory is instantiated and saved internally.

        Plugin writers are expected to override when additional configuration of the
        DownloaderFactory is needed.

        Returns:
            DownloadFactory: The instantiated DownloaderFactory to be used by
                get_downloader()

        """
        try:
            return self._download_factory
        except AttributeError:
            self._download_factory = AnsibleDownloaderFactory(self)
            return self._download_factory

    @hook(
        AFTER_UPDATE,
        when_any=["url", "requirements_file", "sync_dependencies", "signed_only"],
        has_changed=True,
    )
    def _reset_repository_last_synced_metadata_time(self):
        AnsibleRepository.objects.filter(
            remote_id=self.pk, last_synced_metadata_time__isnull=False
        ).update(last_synced_metadata_time=None)

    class Meta:
        default_related_name = "%(app_label)s_%(model_name)s"
        permissions = (("manage_roles_collectionremote", "Can manage roles on collection remotes"),)


class GitRemote(Remote, AutoAddObjPermsMixin):
    """
    A Remote for Collection content hosted in Git repositories.
    """

    TYPE = "git"

    metadata_only = models.BooleanField(default=False)
    git_ref = models.TextField()

    class Meta:
        default_related_name = "%(app_label)s_%(model_name)s"
        permissions = [
            ("manage_roles_gitremote", "Can manage roles on git remotes"),
        ]


class AnsibleCollectionDeprecated(Content):
    """
    A model that represents if a Collection is `deprecated` for a given RepositoryVersion.
    """

    TYPE = "collection_deprecation"

    namespace = models.CharField(max_length=64, editable=False)
    name = models.CharField(max_length=64, editable=False)
    _pulp_domain = models.ForeignKey("core.Domain", default=get_domain_pk, on_delete=models.PROTECT)

    class Meta:
        default_related_name = "%(app_label)s_%(model_name)s"
        unique_together = ("namespace", "name", "_pulp_domain")


class AnsibleRepository(Repository, AutoAddObjPermsMixin):
    """
    Repository for "ansible" content.

    Fields:

        last_synced_metadata_time (models.DateTimeField): Last synced metadata time.
        private (models.BooleanField): Indicator if this repository is private
    """

    TYPE = "ansible"
    CONTENT_TYPES = [
        Role,
        CollectionVersion,
        AnsibleCollectionDeprecated,
        CollectionVersionSignature,
        AnsibleNamespaceMetadata,
        CollectionVersionMark,
    ]
    REMOTE_TYPES = [RoleRemote, CollectionRemote, GitRemote]

    last_synced_metadata_time = models.DateTimeField(null=True)
    gpgkey = models.TextField(null=True)
    private = models.BooleanField(default=False)

    class Meta:
        default_related_name = "%(app_label)s_%(model_name)s"

        permissions = [
            ("rebuild_metadata_ansiblerepository", "Can rebuild metadata on the repository"),
            ("repair_ansiblerepository", "Can repair the repository"),
            ("sign_ansiblerepository", "Can sign content on the repository"),
            ("sync_ansiblerepository", "Can start a sync task on the repository"),
            ("manage_roles_ansiblerepository", "Can manage roles on repositories"),
            ("modify_ansible_repo_content", "Can modify repository content"),
        ]

    def finalize_new_version(self, new_version):
        """Finalize repo version."""
        remove_duplicates(new_version)
        removed_collection_versions = new_version.removed(
            base_version=new_version.base_version
        ).filter(pulp_type=CollectionVersion.get_pulp_type())

        # Remove any deprecated, signature, mark and namespace content
        # associated with the removed collection versions
        for version in removed_collection_versions:
            version = version.cast()

            signatures = new_version.get_content(
                content_qs=CollectionVersionSignature.objects.filter(signed_collection=version)
            )
            new_version.remove_content(signatures)

            marks = new_version.get_content(
                content_qs=CollectionVersionMark.objects.filter(marked_collection=version)
            )
            new_version.remove_content(marks)

            other_collection_versions = new_version.get_content(
                content_qs=CollectionVersion.objects.filter(collection=version.collection)
            )

            # AnsibleCollectionDeprecated and Namespace applies to all collection versions in a
            # repository, so only remove it if there are no more collection versions for the
            # specified collection present.
            if not other_collection_versions.exists():
                deprecations = new_version.get_content(
                    content_qs=AnsibleCollectionDeprecated.objects.filter(
                        namespace=version.namespace, name=version.name
                    )
                )
                namespace = new_version.get_content(
                    content_qs=AnsibleNamespaceMetadata.objects.filter(name=version.namespace)
                )

                new_version.remove_content(deprecations)
                new_version.remove_content(namespace)
        remove_duplicates(new_version)
        validate_repo_version(new_version)

        from pulp_ansible.app.tasks.collectionversion_index import update_index

        update_index(repository_version=new_version, is_latest=True)

    @hook(BEFORE_UPDATE, when="remote", has_changed=True)
    def _reset_repository_last_synced_metadata_time(self):
        self.last_synced_metadata_time = None


class AnsibleDistribution(Distribution, AutoAddObjPermsMixin):
    """
    A Distribution for Ansible content.
    """

    TYPE = "ansible"

    class Meta:
        default_related_name = "%(app_label)s_%(model_name)s"
        permissions = [("manage_roles_ansibledistribution", "Can manage roles on distributions")]

    @hook(AFTER_CREATE)
    @hook(AFTER_UPDATE)
    @hook(AFTER_DELETE)
    def _update_index(self):
        from pulp_ansible.app.tasks.collectionversion_index import update_distribution_index

        update_distribution_index(self)


class CrossRepositoryCollectionVersionIndex(models.Model):
    """
    A model that indexes all CV content across all repositories.
    """

    repository = models.ForeignKey(AnsibleRepository, on_delete=models.CASCADE)
    repository_version = models.ForeignKey(RepositoryVersion, on_delete=models.CASCADE, null=True)
    collection_version = models.ForeignKey(CollectionVersion, on_delete=models.CASCADE)
    namespace_metadata = models.ForeignKey(
        AnsibleNamespaceMetadata, null=True, on_delete=models.SET_NULL
    )

    is_deprecated = models.BooleanField()
    is_signed = models.BooleanField()
    is_highest = models.BooleanField()

    class Meta:
        default_related_name = "%(app_label)s_%(model_name)s"
        unique_together = ("repository", "repository_version", "collection_version")



import fileinput
import math
import os
import random
import shutil
import string
import subprocess
import uuid

from django.db import IntegrityError
from pulpcore.plugin.models import Artifact
from pulpcore.plugin.tasking import dispatch
from pulpcore.plugin.util import get_domain

from pulp_ansible.app.models import AnsibleRepository, CollectionVersion
from pulp_ansible.app.tasks.collections import import_collection


def create_namespace(base_path, collection_n=20, versions_per_collection=5):
    """
    This will generate multiple collections tarballs all with the same namespace.

    This generates `collection_n` collections with random names, and `versions_per_collection`
    number of versions per collection.  By default it makes 20 collections with 5 versions each for
    a total of 100 tarballs with its defaults.

    The versions are 1.0.0, 2.0.0, 3.0.0, ..., etc.

    Args:
        base_path: The file path to write the collections to, e.g. `/home/vagrant/collections`
        collection_n: The number of collections to generate for this namespace
        versions_per_collection: The number of versions per collection

    """
    namespace_name = "".join([random.choice(string.ascii_lowercase) for i in range(26)])
    for collection_i in range(collection_n):
        collection_name = "".join([random.choice(string.ascii_lowercase) for i in range(26)])
        for version_i in range(versions_per_collection):
            version = str(version_i + 1) + ".0.0"

            cmd = "ansible-galaxy collection init --force --init-path {} {}.{}".format(
                base_path, namespace_name, collection_name
            )
            subprocess.run(cmd.split())

            path_to_namespace_dir = os.path.join(base_path, namespace_name)
            path_to_collection_dir = os.path.join(base_path, namespace_name, collection_name)
            galaxy_yml_path = os.path.join(path_to_collection_dir, "galaxy.yml")
            with fileinput.FileInput(galaxy_yml_path, inplace=True) as file:
                for line in file:
                    print(line.replace("1.0.0", version), end="")

            cmd = "ansible-galaxy collection build --output-path {} {}".format(
                base_path, path_to_collection_dir
            )
            subprocess.run(cmd.split())

            shutil.rmtree(path_to_namespace_dir)  # Remove the collection build dir


def import_collection_from_path(path):
    """
    Import a single collection by path.

    This method will not fail if the Artifact already exists.

    Args:
        path: The path to the tarball to import.

    """
    artifact = Artifact.init_and_validate(path)

    try:
        artifact.save()
    except IntegrityError:
        artifact = Artifact.objects.get(sha256=artifact.sha256, pulp_domain=get_domain())

    import_collection(artifact.pk)


def create_repos_with_collections(repo_count, num_repo_versions, collection_percentage):
    """
    Create repositories with a few RepositoryVersions containing existing CollectionVersions.

    This is designed to load a Pulp system with `repo_count` repositories and `num_repo_version` of
    repository versions in each. The content is already existing CollectionVersion objects.

    You specify the percentage of existing CollectionVersion objects using the
    `collection_percentage` argument. This needs to be between 0.0 and 1.0, e.g. 0.9 will select a
    random 90% of all CollectionVersions in each RepositoryVersion it creates.

    Args:
        repo_count: The number of repositories to make.
        num_repo_versions: The number of RepsitoryVersion objects per repository to make.
        collection_percentage: The float between 0.0 and 1.0 of the percentage of CollectionVersion
            objects to use in each RepositoryVersion. For example, 0.9 would have a random 90% of
            CollectionVersion objects in each RepositoryVersion it creates.

    """
    content_pk = [i for i in CollectionVersion.objects.values_list("pk", flat=True)]
    last_content_index = math.floor(collection_percentage * len(content_pk))

    for repository_i in range(repo_count):
        repository, created = AnsibleRepository.objects.get_or_create(name=uuid.uuid4())

        for version_i in range(num_repo_versions):
            random.shuffle(content_pk)
            with repository.new_version() as new_version:
                qs = CollectionVersion.objects.filter(pk__in=content_pk[:last_content_index])
                new_version.add_content(qs)


def promote_content(repos_per_task, num_repos_to_update):
    """
    Select a random CollectionVersion and attempt to add it to some number of AnsibleRepositories.

    By default this will update all Repositories, creating a new RepositoryVersion on each. The
    `num_repos_to_update` argument can specify the number of Repositories to update.

    This task generates many subtasks. This task randomly selects the Repositories it will update,
    and randomly selects the CollectionVersion. Then it dispatches N repositories to be handled by
    the `add_content_to_repositories` task. The dispatch locks on all repositories in the set to
    make this workload safe for execution with any other Pulp workload.

    The `repos_per_task` argument controls the number of repositories handled per subtask.

    Args:
        repos_per_task: The number of repositories to handle in each subtask. 100 is a typical
            choice.
        num_repos_to_update: The total number of repositories to update.

    """
    random_collection_version = (
        CollectionVersion.objects.order_by("?").filter(pulp_domain=get_domain()).first()
    )
    repos_to_dispatch = []
    locks = []
    for repo_num, repo in enumerate(AnsibleRepository.objects.filter(pulp_domain=get_domain()), 1):
        if repo_num > num_repos_to_update:
            break
        repos_to_dispatch.append(repo.pk)
        locks.append(repo)
        if len(repos_to_dispatch) == repos_per_task:
            task_args = (random_collection_version.pk, repos_to_dispatch)
            dispatch(add_content_to_repositories, exclusive_resources=locks, args=task_args)
            repos_to_dispatch = []
            locks = []

    if repos_to_dispatch:
        task_args = (random_collection_version.pk, repos_to_dispatch)
        dispatch(add_content_to_repositories, exclusive_resources=locks, args=task_args)


def add_content_to_repositories(collection_version_pk, repositories_pks):
    """
    Add a CollectionVersion to many repositories.

    Args:
        collection_version_pk: The pk of the CollectionVersion to add to each repository.
        repositories_pks: The pks of the AnsibleRepository to add the CollectionVersion to.

    """
    for repository in AnsibleRepository.objects.filter(pk__in=repositories_pks):
        with repository.new_version() as new_version:
            qs = CollectionVersion.objects.filter(pk=collection_version_pk)
            new_version.add_content(qs)
